# -*- coding: utf-8 -*-

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import io
import os
import pytest
import textwrap
from unittest import mock

import docker

from google.cloud.aiplatform.constants import prediction
from google.cloud.aiplatform.docker_utils import build
from google.cloud.aiplatform.docker_utils import errors
from google.cloud.aiplatform.docker_utils import local_util
from google.cloud.aiplatform.docker_utils import run
from google.cloud.aiplatform.docker_utils import utils
from google.cloud.aiplatform.utils import prediction_utils


_TEST_CONTAINER_LOGS = b"line1\nline2\nline3\n"
_TEST_CONTAINER_LOGS_LEN = 3


@pytest.fixture
def docker_client_mock():
    with mock.patch("docker.from_env") as from_env_mock:
        client = from_env_mock.return_value
        client().containers.run.return_value = None
        client.images.get.return_value = None
        yield client


@pytest.fixture
def docker_client_mock_image_get_not_found():
    with mock.patch("docker.from_env") as from_env_mock:
        client = from_env_mock.return_value
        client.images.get.side_effect = docker.errors.ImageNotFound("")
        yield client


@pytest.fixture
def docker_client_mock_image_get_api_error():
    with mock.patch("docker.from_env") as from_env_mock:
        client = from_env_mock.return_value
        client.images.get.side_effect = docker.errors.APIError("")
        yield client


@pytest.fixture
def docker_container_mock():
    container = mock.MagicMock()
    container.logs.return_value = _TEST_CONTAINER_LOGS
    return container


@pytest.fixture
def execute_command_mock():
    with mock.patch.object(local_util, "execute_command") as execute_command_mock:
        execute_command_mock.return_value = 0
        yield execute_command_mock


@pytest.fixture
def execute_command_return_code_1_mock():
    with mock.patch.object(
        local_util, "execute_command"
    ) as execute_command_return_code_1_mock:
        execute_command_return_code_1_mock.return_value = 1
        yield execute_command_return_code_1_mock


class MockedPopen:
    def __init__(self, args, **kwargs):
        self.args = args
        self.returncode = 0
        self.stdin = mock.Mock()
        self.stdout = mock.Mock(return_value="fake output")

    def __enter__(self):
        return self

    def __exit__(self, exe_type, value, traceback):
        pass

    def communicate(self, input=None, timeout=None):
        return "", ""


@pytest.fixture
def textiowrapper_mock():
    with mock.patch.object(io, "TextIOWrapper") as textiowrapper_mock:
        textiowrapper_mock.return_value = io.TextIOWrapper(io.BytesIO(b""))
        yield textiowrapper_mock


@pytest.fixture
def make_dockerfile_mock():
    with mock.patch.object(build, "make_dockerfile") as make_dockerfile_mock:
        yield make_dockerfile_mock


class TestRun:
    IMAGE_URI = "test_image:latest"

    @mock.patch.dict(os.environ, {}, clear=True)
    def test_run_prediction_container(self, docker_client_mock):
        run.run_prediction_container(self.IMAGE_URI)

        docker_client_mock.containers.run.assert_called_once_with(
            self.IMAGE_URI,
            command=None,
            entrypoint=None,
            ports={prediction.DEFAULT_AIP_HTTP_PORT: None},
            environment={
                prediction.AIP_HTTP_PORT: prediction.DEFAULT_AIP_HTTP_PORT,
                prediction.AIP_HEALTH_ROUTE: None,
                prediction.AIP_PREDICT_ROUTE: None,
                prediction.AIP_STORAGE_URI: "",
            },
            volumes=[],
            device_requests=None,
            detach=True,
        )

    def test_run_prediction_container_with_all_parameters(
        self, tmp_path, docker_client_mock
    ):
        artifact_uri = "gs://myproject/mymodel"
        serving_container_predict_route = "/custom_predict"
        serving_container_health_route = "/custom_health"
        serving_container_command = ["echo", "hello"]
        serving_container_args = [">", "tmp.log"]
        serving_container_environment_variables = {"custom_key": "custom_value"}
        serving_container_ports = [5555]
        credential_path = tmp_path / "key.json"
        credential_path.write_text("")
        host_port = 6666
        environment = {k: v for k, v in serving_container_environment_variables.items()}
        environment[prediction.AIP_HTTP_PORT] = serving_container_ports[0]
        environment[prediction.AIP_HEALTH_ROUTE] = serving_container_health_route
        environment[prediction.AIP_PREDICT_ROUTE] = serving_container_predict_route
        environment[prediction.AIP_STORAGE_URI] = artifact_uri
        environment[
            run._ADC_ENVIRONMENT_VARIABLE
        ] = run._DEFAULT_CONTAINER_CRED_KEY_PATH
        volumes = [f"{credential_path}:{run._DEFAULT_CONTAINER_CRED_KEY_PATH}"]

        run.run_prediction_container(
            self.IMAGE_URI,
            artifact_uri=artifact_uri,
            serving_container_predict_route=serving_container_predict_route,
            serving_container_health_route=serving_container_health_route,
            serving_container_command=serving_container_command,
            serving_container_args=serving_container_args,
            serving_container_environment_variables=serving_container_environment_variables,
            serving_container_ports=serving_container_ports,
            credential_path=credential_path,
            host_port=host_port,
        )

        docker_client_mock.containers.run.assert_called_once_with(
            self.IMAGE_URI,
            command=serving_container_args,
            entrypoint=serving_container_command,
            ports={serving_container_ports[0]: host_port},
            environment=environment,
            volumes=volumes,
            device_requests=None,
            detach=True,
        )

    @mock.patch.dict(os.environ, {}, clear=True)
    def test_run_prediction_container_with_envs_replaced_by_envs(
        self, tmp_path, docker_client_mock
    ):
        serving_container_environment_variables = {
            "VAR_1": "foo",
            "VAR_2": "$(VAR_3) bar",
            "VAR_3": "$(VAR_1) bar",
            "VAR_4": "$$(VAR_1)",
        }
        environment = {k: v for k, v in serving_container_environment_variables.items()}
        environment[prediction.AIP_HTTP_PORT] = prediction.DEFAULT_AIP_HTTP_PORT
        environment[prediction.AIP_HEALTH_ROUTE] = None
        environment[prediction.AIP_PREDICT_ROUTE] = None
        environment[prediction.AIP_STORAGE_URI] = ""
        # Envs referencing earlier entries will be changed. Those envs referencing later
        # entries won't be changed.
        environment["VAR_3"] = "foo bar"
        # Double $$ will be replaced with a single $.
        environment["VAR_4"] = "$(VAR_1)"

        run.run_prediction_container(
            self.IMAGE_URI,
            serving_container_environment_variables=serving_container_environment_variables,
        )

        docker_client_mock.containers.run.assert_called_once_with(
            self.IMAGE_URI,
            command=None,
            entrypoint=None,
            ports={prediction.DEFAULT_AIP_HTTP_PORT: None},
            environment=environment,
            volumes=[],
            device_requests=None,
            detach=True,
        )

    @mock.patch.dict(os.environ, {}, clear=True)
    def test_run_prediction_container_with_command_replaced_by_envs(
        self, tmp_path, docker_client_mock
    ):
        serving_container_command = ["$(VAR_1)", "$$(VAR_1)", "$(VAR_2)"]
        serving_container_environment_variables = {
            "VAR_1": "foo",
        }
        environment = {k: v for k, v in serving_container_environment_variables.items()}
        environment[prediction.AIP_HTTP_PORT] = prediction.DEFAULT_AIP_HTTP_PORT
        environment[prediction.AIP_HEALTH_ROUTE] = None
        environment[prediction.AIP_PREDICT_ROUTE] = None
        environment[prediction.AIP_STORAGE_URI] = ""
        # Command references existing environment variables.
        expected_entrypoint = ["foo", "$(VAR_1)", "$(VAR_2)"]

        run.run_prediction_container(
            self.IMAGE_URI,
            serving_container_command=serving_container_command,
            serving_container_environment_variables=serving_container_environment_variables,
        )

        docker_client_mock.containers.run.assert_called_once_with(
            self.IMAGE_URI,
            command=None,
            entrypoint=expected_entrypoint,
            ports={prediction.DEFAULT_AIP_HTTP_PORT: None},
            environment=environment,
            volumes=[],
            device_requests=None,
            detach=True,
        )

    @mock.patch.dict(os.environ, {}, clear=True)
    def test_run_prediction_container_with_args_replaced_by_envs(
        self, tmp_path, docker_client_mock
    ):
        serving_container_args = ["$(VAR_1)", "$$(VAR_1)", "$(VAR_2)"]
        serving_container_environment_variables = {
            "VAR_1": "foo",
        }
        environment = {k: v for k, v in serving_container_environment_variables.items()}
        environment[prediction.AIP_HTTP_PORT] = prediction.DEFAULT_AIP_HTTP_PORT
        environment[prediction.AIP_HEALTH_ROUTE] = None
        environment[prediction.AIP_PREDICT_ROUTE] = None
        environment[prediction.AIP_STORAGE_URI] = ""
        # Args references existing environment variables.
        expected_command = ["foo", "$(VAR_1)", "$(VAR_2)"]

        run.run_prediction_container(
            self.IMAGE_URI,
            serving_container_args=serving_container_args,
            serving_container_environment_variables=serving_container_environment_variables,
        )

        docker_client_mock.containers.run.assert_called_once_with(
            self.IMAGE_URI,
            command=expected_command,
            entrypoint=None,
            ports={prediction.DEFAULT_AIP_HTTP_PORT: None},
            environment=environment,
            volumes=[],
            device_requests=None,
            detach=True,
        )

    def test_run_prediction_container_credential_from_adc(
        self, tmp_path, docker_client_mock
    ):
        credential_path = tmp_path / "key.json"
        credential_path.write_text("")
        volumes = [f"{credential_path}:{run._DEFAULT_CONTAINER_CRED_KEY_PATH}"]

        with mock.patch.dict(
            os.environ, {run._ADC_ENVIRONMENT_VARIABLE: credential_path.as_posix()}
        ):
            run.run_prediction_container(self.IMAGE_URI)

        docker_client_mock.containers.run.assert_called_once_with(
            self.IMAGE_URI,
            command=None,
            entrypoint=None,
            ports={prediction.DEFAULT_AIP_HTTP_PORT: None},
            environment={
                prediction.AIP_HTTP_PORT: prediction.DEFAULT_AIP_HTTP_PORT,
                prediction.AIP_HEALTH_ROUTE: None,
                prediction.AIP_PREDICT_ROUTE: None,
                prediction.AIP_STORAGE_URI: "",
                run._ADC_ENVIRONMENT_VARIABLE: run._DEFAULT_CONTAINER_CRED_KEY_PATH,
            },
            volumes=volumes,
            device_requests=None,
            detach=True,
        )

    @mock.patch.dict(os.environ, {}, clear=True)
    def test_run_prediction_container_gpu_count(self, docker_client_mock):
        gpu_count = 1
        gpu_capabilities = [["gpu"]]

        run.run_prediction_container(
            self.IMAGE_URI,
            gpu_count=gpu_count,
            gpu_capabilities=gpu_capabilities,
        )

        docker_client_mock.containers.run.assert_called_once_with(
            self.IMAGE_URI,
            command=None,
            entrypoint=None,
            ports={prediction.DEFAULT_AIP_HTTP_PORT: None},
            environment={
                prediction.AIP_HTTP_PORT: prediction.DEFAULT_AIP_HTTP_PORT,
                prediction.AIP_HEALTH_ROUTE: None,
                prediction.AIP_PREDICT_ROUTE: None,
                prediction.AIP_STORAGE_URI: "",
            },
            volumes=[],
            device_requests=[
                docker.types.DeviceRequest(
                    count=gpu_count, capabilities=gpu_capabilities
                )
            ],
            detach=True,
        )

    @mock.patch.dict(os.environ, {}, clear=True)
    def test_run_prediction_container_gpu_device_ids(self, docker_client_mock):
        gpu_device_ids = ["1"]
        gpu_capabilities = [["gpu"]]

        run.run_prediction_container(
            self.IMAGE_URI,
            gpu_device_ids=gpu_device_ids,
            gpu_capabilities=gpu_capabilities,
        )

        docker_client_mock.containers.run.assert_called_once_with(
            self.IMAGE_URI,
            command=None,
            entrypoint=None,
            ports={prediction.DEFAULT_AIP_HTTP_PORT: None},
            environment={
                prediction.AIP_HTTP_PORT: prediction.DEFAULT_AIP_HTTP_PORT,
                prediction.AIP_HEALTH_ROUTE: None,
                prediction.AIP_PREDICT_ROUTE: None,
                prediction.AIP_STORAGE_URI: "",
            },
            volumes=[],
            device_requests=[
                docker.types.DeviceRequest(
                    device_ids=gpu_device_ids, capabilities=gpu_capabilities
                )
            ],
            detach=True,
        )

    def test_run_prediction_container_artifact_uri_is_local_path_default_workdir(
        self, tmp_path, docker_client_mock
    ):
        artifact_uri = tmp_path / "models"
        artifact_uri.mkdir()
        fake_model_artifact = artifact_uri / "model.pb"
        fake_model_artifact.write_text("")
        environment = {}
        environment[prediction.AIP_HTTP_PORT] = prediction.DEFAULT_AIP_HTTP_PORT
        environment[prediction.AIP_HEALTH_ROUTE] = None
        environment[prediction.AIP_PREDICT_ROUTE] = None
        environment[prediction.AIP_STORAGE_URI] = utils.DEFAULT_MOUNTED_MODEL_DIRECTORY
        environment[
            run._ADC_ENVIRONMENT_VARIABLE
        ] = run._DEFAULT_CONTAINER_CRED_KEY_PATH
        credential_path = tmp_path / "key.json"
        credential_path.write_text("")
        volumes = [
            f"{fake_model_artifact.as_posix()}:{utils.DEFAULT_MOUNTED_MODEL_DIRECTORY + '/model.pb'}",
            f"{credential_path}:{run._DEFAULT_CONTAINER_CRED_KEY_PATH}",
        ]

        with mock.patch.dict(
            os.environ, {run._ADC_ENVIRONMENT_VARIABLE: credential_path.as_posix()}
        ):
            run.run_prediction_container(
                self.IMAGE_URI,
                artifact_uri=artifact_uri.as_posix(),
            )

        docker_client_mock.containers.run.assert_called_once_with(
            self.IMAGE_URI,
            command=None,
            entrypoint=None,
            ports={prediction.DEFAULT_AIP_HTTP_PORT: None},
            environment=environment,
            volumes=volumes,
            device_requests=None,
            detach=True,
        )

    def test_run_prediction_container_artifact_uri_is_local_path_but_not_exists(
        self, tmp_path, docker_client_mock
    ):
        artifact_uri = tmp_path / "models"
        environment = {}
        environment[prediction.AIP_HTTP_PORT] = prediction.DEFAULT_AIP_HTTP_PORT
        environment[prediction.AIP_HEALTH_ROUTE] = None
        environment[prediction.AIP_PREDICT_ROUTE] = None
        environment[prediction.AIP_STORAGE_URI] = utils.DEFAULT_WORKDIR
        environment[
            run._ADC_ENVIRONMENT_VARIABLE
        ] = run._DEFAULT_CONTAINER_CRED_KEY_PATH
        credential_path = tmp_path / "key.json"
        credential_path.write_text("")
        expected_message = (
            "artifact_uri should be specified as either a GCS uri which starts with "
            f"`{prediction_utils.GCS_URI_PREFIX}` or a path to a local directory. "
            f'However, "{artifact_uri}" does not exist.'
        )

        with pytest.raises(ValueError) as exception:
            with mock.patch.dict(
                os.environ, {run._ADC_ENVIRONMENT_VARIABLE: credential_path.as_posix()}
            ):
                run.run_prediction_container(
                    self.IMAGE_URI,
                    artifact_uri=artifact_uri.as_posix(),
                )

        assert str(exception.value) == expected_message

    def test_run_prediction_container_credential_path_not_exists(
        self, docker_client_mock
    ):
        credential_path = "key.json"
        expected_message = f'credential_path does not exist: "{credential_path}".'

        with pytest.raises(ValueError) as exception:
            run.run_prediction_container(
                self.IMAGE_URI, credential_path=credential_path
            )

        assert str(exception.value) == expected_message

    @mock.patch.dict(os.environ, {run._ADC_ENVIRONMENT_VARIABLE: "key.json"})
    def test_run_prediction_container_adc_value_not_exists(self, docker_client_mock):
        expected_message = (
            f"The file from the environment variable {run._ADC_ENVIRONMENT_VARIABLE} does "
            f'not exist: "key.json".'
        )

        with pytest.raises(ValueError) as exception:
            run.run_prediction_container(self.IMAGE_URI)

        assert str(exception.value) == expected_message

    def test_print_container_logs(self, docker_container_mock):
        with mock.patch(
            "google.cloud.aiplatform.docker_utils.run._logger"
        ) as logger_mock:
            logs_len = run.print_container_logs(docker_container_mock)

        assert logs_len == _TEST_CONTAINER_LOGS_LEN
        assert docker_container_mock.logs.called
        assert logger_mock.info.call_count == _TEST_CONTAINER_LOGS_LEN

    def test_print_container_logs_with_start_index(self, docker_container_mock):
        start_index = 1
        with mock.patch(
            "google.cloud.aiplatform.docker_utils.run._logger"
        ) as logger_mock:
            logs_len = run.print_container_logs(
                docker_container_mock, start_index=start_index
            )

        assert logs_len == _TEST_CONTAINER_LOGS_LEN
        assert docker_container_mock.logs.called
        assert logger_mock.info.call_count == (_TEST_CONTAINER_LOGS_LEN - start_index)

    def test_print_container_logs_with_message(self, docker_container_mock):
        with mock.patch(
            "google.cloud.aiplatform.docker_utils.run._logger"
        ) as logger_mock:
            logs_len = run.print_container_logs(
                docker_container_mock, message="Test message:"
            )

        assert logs_len == _TEST_CONTAINER_LOGS_LEN
        assert docker_container_mock.logs.called
        assert logger_mock.info.call_count == _TEST_CONTAINER_LOGS_LEN + 1


class TestBuild:
    BASE_IMAGE = "python:3.7"
    SOURCE_DIR = "src"
    HOST_WORKDIR_BASENAME = "user_code"
    HOST_WORKDIR = f"./{SOURCE_DIR}/{HOST_WORKDIR_BASENAME}"
    HOME = utils.DEFAULT_HOME
    WORKDIR = utils.DEFAULT_WORKDIR
    SCRIPT = "./user_code/entrypoint.py"
    SCRIPT_PACKAGE_PATH = "user_code/entrypoint.py"
    PYTHON_MODULE = "custom.python.module"
    PACKAGE = utils.Package(
        script=SCRIPT, package_path=HOST_WORKDIR, python_module=None
    )
    PACKAGE_WITH_PYTHON_MODULE = utils.Package(
        script=SCRIPT, package_path=HOST_WORKDIR, python_module=PYTHON_MODULE
    )
    PACKAGE_NO_SCRIPT_AND_MODULE = utils.Package(
        script=None, package_path=HOST_WORKDIR, python_module=None
    )
    OUTPUT_IMAGE_NAME = "test_image:latest"
    REQUIREMENTS_FILE = "requirements.txt"
    EXTRA_PACKAGE = "custom_package.tar.gz"
    SETUP_FILE = "setup.py"
    PIP = "pip"
    PYTHON = "python"

    def test_make_dockerfile(self):
        result = build.make_dockerfile(
            self.BASE_IMAGE, self.PACKAGE, self.WORKDIR, self.HOME
        )

        assert f"FROM {self.BASE_IMAGE}\n" in result
        assert f"WORKDIR {self.WORKDIR}\n" in result
        assert f"ENV HOME={self.HOME}\n" in result
        assert 'COPY [".", "."]\n' in result
        assert f'ENTRYPOINT ["python", "{self.SCRIPT}"]' in result

    def test_make_dockerfile_with_python_module(self):
        result = build.make_dockerfile(
            self.BASE_IMAGE, self.PACKAGE_WITH_PYTHON_MODULE, self.WORKDIR, self.HOME
        )

        assert f"FROM {self.BASE_IMAGE}\n" in result
        assert f"WORKDIR {self.WORKDIR}\n" in result
        assert f"ENV HOME={self.HOME}\n" in result
        assert 'COPY [".", "."]\n' in result
        assert f'ENTRYPOINT ["python", "-m", "{self.PYTHON_MODULE}"]' in result

    def test_make_dockerfile_no_script_and_module(self):
        result = build.make_dockerfile(
            self.BASE_IMAGE, self.PACKAGE_NO_SCRIPT_AND_MODULE, self.WORKDIR, self.HOME
        )

        assert f"FROM {self.BASE_IMAGE}\n" in result
        assert f"WORKDIR {self.WORKDIR}\n" in result
        assert f"ENV HOME={self.HOME}\n" in result
        assert 'COPY [".", "."]\n' in result
        assert "ENTRYPOINT" not in result

    def test_make_dockerfile_with_requirements_path(self):
        requirements_path = "./requirements.txt"

        result = build.make_dockerfile(
            self.BASE_IMAGE,
            self.PACKAGE,
            self.WORKDIR,
            self.HOME,
            requirements_path=requirements_path,
        )

        assert f"FROM {self.BASE_IMAGE}\n" in result
        assert f"WORKDIR {self.WORKDIR}\n" in result
        assert f"ENV HOME={self.HOME}\n" in result
        assert 'COPY [".", "."]\n' in result
        assert f'ENTRYPOINT ["python", "{self.SCRIPT}"]' in result
        assert (
            f"RUN pip install --no-cache-dir --force-reinstall -r {requirements_path}\n"
            in result
        )

    def test_make_dockerfile_with_setup_path(self):
        setup_path = "./custom_setup.py"

        result = build.make_dockerfile(
            self.BASE_IMAGE,
            self.PACKAGE,
            self.WORKDIR,
            self.HOME,
            setup_path=setup_path,
        )

        assert f"FROM {self.BASE_IMAGE}\n" in result
        assert f"WORKDIR {self.WORKDIR}\n" in result
        assert f"ENV HOME={self.HOME}\n" in result
        assert 'COPY [".", "."]\n' in result
        assert f'ENTRYPOINT ["python", "{self.SCRIPT}"]' in result
        assert f'COPY ["{setup_path}", "./setup.py"]\n' in result
        assert "RUN pip install --no-cache-dir --force-reinstall .\n" in result

    def test_make_dockerfile_with_extra_requirements(self):
        extra_requirement = "custom_package==1.0"

        result = build.make_dockerfile(
            self.BASE_IMAGE,
            self.PACKAGE,
            self.WORKDIR,
            self.HOME,
            extra_requirements=[extra_requirement],
        )

        assert f"FROM {self.BASE_IMAGE}\n" in result
        assert f"WORKDIR {self.WORKDIR}\n" in result
        assert f"ENV HOME={self.HOME}\n" in result
        assert 'COPY [".", "."]\n' in result
        assert f'ENTRYPOINT ["python", "{self.SCRIPT}"]' in result
        assert (
            f"RUN pip install --no-cache-dir --force-reinstall {extra_requirement}\n"
            in result
        )

    def test_make_dockerfile_with_extra_packages(self):
        extra_package_basename = "custom_package"
        extra_package = f"./{extra_package_basename}"

        result = build.make_dockerfile(
            self.BASE_IMAGE,
            self.PACKAGE,
            self.WORKDIR,
            self.HOME,
            extra_packages=[extra_package],
        )

        assert f"FROM {self.BASE_IMAGE}\n" in result
        assert f"WORKDIR {self.WORKDIR}\n" in result
        assert f"ENV HOME={self.HOME}\n" in result
        assert 'COPY [".", "."]\n' in result
        assert f'ENTRYPOINT ["python", "{self.SCRIPT}"]' in result
        assert (
            f"RUN pip install --no-cache-dir --force-reinstall {extra_package}\n"
            in result
        )

    def test_make_dockerfile_with_extra_dirs(self):
        extra_dir = "./subdir"

        result = build.make_dockerfile(
            self.BASE_IMAGE,
            self.PACKAGE,
            self.WORKDIR,
            self.HOME,
            extra_dirs=[extra_dir],
        )

        assert f"FROM {self.BASE_IMAGE}\n" in result
        assert f"WORKDIR {self.WORKDIR}\n" in result
        assert f"ENV HOME={self.HOME}\n" in result
        assert 'COPY [".", "."]\n' in result
        assert f'ENTRYPOINT ["python", "{self.SCRIPT}"]' in result
        assert f'COPY ["{extra_dir}", "{extra_dir}"]\n' in result

    def test_make_dockerfile_with_exposed_ports(self):
        exposed_port = 8080

        result = build.make_dockerfile(
            self.BASE_IMAGE,
            self.PACKAGE,
            self.WORKDIR,
            self.HOME,
            exposed_ports=[exposed_port],
        )

        assert f"FROM {self.BASE_IMAGE}\n" in result
        assert f"WORKDIR {self.WORKDIR}\n" in result
        assert f"ENV HOME={self.HOME}\n" in result
        assert 'COPY [".", "."]\n' in result
        assert f'ENTRYPOINT ["python", "{self.SCRIPT}"]' in result
        assert f"EXPOSE {exposed_port}\n" in result

    def test_make_dockerfile_with_environment_variables(self):
        environment_variables = {
            "FAKE_ENV1": "FAKE_VALUE1",
            "FAKE_ENV2": "FAKE_VALUE2",
        }

        result = build.make_dockerfile(
            self.BASE_IMAGE,
            self.PACKAGE,
            self.WORKDIR,
            self.HOME,
            environment_variables=environment_variables,
        )

        assert f"FROM {self.BASE_IMAGE}\n" in result
        assert f"WORKDIR {self.WORKDIR}\n" in result
        assert f"ENV HOME={self.HOME}\n" in result
        assert 'COPY [".", "."]\n' in result
        assert f'ENTRYPOINT ["python", "{self.SCRIPT}"]' in result
        assert "ENV FAKE_ENV1=FAKE_VALUE1\n" in result
        assert "ENV FAKE_ENV2=FAKE_VALUE2\n" in result

    def test_build_image(self, make_dockerfile_mock, execute_command_mock):
        image = build.build_image(
            self.BASE_IMAGE, self.HOST_WORKDIR, self.OUTPUT_IMAGE_NAME
        )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=self.HOST_WORKDIR,
                python_module=None,
            ),
            utils.DEFAULT_WORKDIR,
            utils.DEFAULT_HOME,
            requirements_path=None,
            setup_path=None,
            extra_requirements=None,
            extra_packages=None,
            extra_dirs=None,
            exposed_ports=None,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_mock.assert_called_once_with(
            [
                "docker",
                "build",
                "--no-cache",
                "-t",
                self.OUTPUT_IMAGE_NAME,
                "--rm",
                "-f-",
                self.HOST_WORKDIR,
            ],
            input_str=make_dockerfile_mock.return_value,
        )
        assert image.name == self.OUTPUT_IMAGE_NAME
        assert image.default_home == self.HOME
        assert image.default_workdir == self.WORKDIR

    def test_build_image_with_python_module(
        self, make_dockerfile_mock, execute_command_mock
    ):
        image = build.build_image(
            self.BASE_IMAGE,
            self.HOST_WORKDIR,
            self.OUTPUT_IMAGE_NAME,
            python_module=self.PYTHON_MODULE,
        )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=self.HOST_WORKDIR,
                python_module=self.PYTHON_MODULE,
            ),
            utils.DEFAULT_WORKDIR,
            utils.DEFAULT_HOME,
            requirements_path=None,
            setup_path=None,
            extra_requirements=None,
            extra_packages=None,
            extra_dirs=None,
            exposed_ports=None,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_mock.assert_called_once_with(
            [
                "docker",
                "build",
                "--no-cache",
                "-t",
                self.OUTPUT_IMAGE_NAME,
                "--rm",
                "-f-",
                self.HOST_WORKDIR,
            ],
            input_str=make_dockerfile_mock.return_value,
        )
        assert image.name == self.OUTPUT_IMAGE_NAME
        assert image.default_home == self.HOME
        assert image.default_workdir == self.WORKDIR

    def test_build_image_with_extra_requirements(
        self, make_dockerfile_mock, execute_command_mock
    ):
        extra_requirements = ["custom_package==1.0"]
        image = build.build_image(
            self.BASE_IMAGE,
            self.HOST_WORKDIR,
            self.OUTPUT_IMAGE_NAME,
            extra_requirements=extra_requirements,
        )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=self.HOST_WORKDIR,
                python_module=None,
            ),
            utils.DEFAULT_WORKDIR,
            utils.DEFAULT_HOME,
            requirements_path=None,
            setup_path=None,
            extra_requirements=extra_requirements,
            extra_packages=None,
            extra_dirs=None,
            exposed_ports=None,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_mock.assert_called_once_with(
            [
                "docker",
                "build",
                "--no-cache",
                "-t",
                self.OUTPUT_IMAGE_NAME,
                "--rm",
                "-f-",
                self.HOST_WORKDIR,
            ],
            input_str=make_dockerfile_mock.return_value,
        )
        assert image.name == self.OUTPUT_IMAGE_NAME
        assert image.default_home == self.HOME
        assert image.default_workdir == self.WORKDIR

    def test_build_image_with_requirements_path(
        self, tmp_path, make_dockerfile_mock, execute_command_mock
    ):
        source_dir = tmp_path / self.SOURCE_DIR
        source_dir.mkdir()
        host_workdir = tmp_path / self.HOST_WORKDIR
        host_workdir.mkdir()
        requirements_file = host_workdir / self.REQUIREMENTS_FILE
        requirements_file.write_text("")

        image = build.build_image(
            self.BASE_IMAGE,
            host_workdir.as_posix(),
            self.OUTPUT_IMAGE_NAME,
            requirements_path=requirements_file.as_posix(),
        )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=host_workdir.as_posix(),
                python_module=None,
            ),
            utils.DEFAULT_WORKDIR,
            utils.DEFAULT_HOME,
            requirements_path=self.REQUIREMENTS_FILE,
            setup_path=None,
            extra_requirements=None,
            extra_packages=None,
            extra_dirs=None,
            exposed_ports=None,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_mock.assert_called_once_with(
            [
                "docker",
                "build",
                "--no-cache",
                "-t",
                self.OUTPUT_IMAGE_NAME,
                "--rm",
                "-f-",
                host_workdir.as_posix(),
            ],
            input_str=make_dockerfile_mock.return_value,
        )
        assert image.name == self.OUTPUT_IMAGE_NAME
        assert image.default_home == self.HOME
        assert image.default_workdir == self.WORKDIR

    def test_build_image_not_found_requirements_path(self, make_dockerfile_mock):
        requirements_path = f"./another_src/{self.REQUIREMENTS_FILE}"
        expected_message = f'The requirements_path "{requirements_path}" must exist.'

        with pytest.raises(ValueError) as exception:
            _ = build.build_image(
                self.BASE_IMAGE,
                self.HOST_WORKDIR,
                self.OUTPUT_IMAGE_NAME,
                requirements_path=requirements_path,
            )

        assert not make_dockerfile_mock.called
        assert str(exception.value) == expected_message

    def test_build_image_invalid_requirements_path(
        self, tmp_path, make_dockerfile_mock
    ):
        source_dir = tmp_path / self.SOURCE_DIR
        source_dir.mkdir()
        host_workdir = tmp_path / self.HOST_WORKDIR
        host_workdir.mkdir()
        another_dir = tmp_path / "another_dir"
        another_dir.mkdir()
        requirements_file = another_dir / self.REQUIREMENTS_FILE
        requirements_file.write_text("")
        expected_message = (
            f'The requirements_path "{requirements_file}" must be in "{host_workdir}".'
        )

        with pytest.raises(ValueError) as exception:
            _ = build.build_image(
                self.BASE_IMAGE,
                host_workdir.as_posix(),
                self.OUTPUT_IMAGE_NAME,
                requirements_path=requirements_file.as_posix(),
            )

        assert not make_dockerfile_mock.called
        assert str(exception.value) == expected_message

    def test_build_image_with_setup_path(
        self, tmp_path, make_dockerfile_mock, execute_command_mock
    ):
        source_dir = tmp_path / self.SOURCE_DIR
        source_dir.mkdir()
        host_workdir = tmp_path / self.HOST_WORKDIR
        host_workdir.mkdir()
        setup_file = host_workdir / self.SETUP_FILE
        setup_file.write_text("")

        image = build.build_image(
            self.BASE_IMAGE,
            host_workdir.as_posix(),
            self.OUTPUT_IMAGE_NAME,
            setup_path=setup_file.as_posix(),
        )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=host_workdir.as_posix(),
                python_module=None,
            ),
            utils.DEFAULT_WORKDIR,
            utils.DEFAULT_HOME,
            requirements_path=None,
            setup_path=self.SETUP_FILE,
            extra_requirements=None,
            extra_packages=None,
            extra_dirs=None,
            exposed_ports=None,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_mock.assert_called_once_with(
            [
                "docker",
                "build",
                "--no-cache",
                "-t",
                self.OUTPUT_IMAGE_NAME,
                "--rm",
                "-f-",
                host_workdir.as_posix(),
            ],
            input_str=make_dockerfile_mock.return_value,
        )
        assert image.name == self.OUTPUT_IMAGE_NAME
        assert image.default_home == self.HOME
        assert image.default_workdir == self.WORKDIR

    def test_build_image_not_found_setup_path(self, make_dockerfile_mock):
        setup_path = f"./another_src/{self.SETUP_FILE}"
        expected_message = f'The setup_path "{setup_path}" must exist.'

        with pytest.raises(ValueError) as exception:
            _ = build.build_image(
                self.BASE_IMAGE,
                self.HOST_WORKDIR,
                self.OUTPUT_IMAGE_NAME,
                setup_path=setup_path,
            )

        assert not make_dockerfile_mock.called
        assert str(exception.value) == expected_message

    def test_build_image_invalid_setup_path(self, tmp_path, make_dockerfile_mock):
        source_dir = tmp_path / self.SOURCE_DIR
        source_dir.mkdir()
        host_workdir = tmp_path / self.HOST_WORKDIR
        host_workdir.mkdir()
        another_dir = tmp_path / "another_dir"
        another_dir.mkdir()
        setup_file = another_dir / self.SETUP_FILE
        setup_file.write_text("")
        expected_message = f'The setup_path "{setup_file}" must be in "{host_workdir}".'

        with pytest.raises(ValueError) as exception:
            _ = build.build_image(
                self.BASE_IMAGE,
                host_workdir.as_posix(),
                self.OUTPUT_IMAGE_NAME,
                setup_path=setup_file.as_posix(),
            )

        assert not make_dockerfile_mock.called
        assert str(exception.value) == expected_message

    def test_build_image_with_extra_packages(
        self, tmp_path, make_dockerfile_mock, execute_command_mock
    ):
        source_dir = tmp_path / self.SOURCE_DIR
        source_dir.mkdir()
        host_workdir = tmp_path / self.HOST_WORKDIR
        host_workdir.mkdir()
        extra_package = host_workdir / self.EXTRA_PACKAGE
        extra_package.write_text("")

        image = build.build_image(
            self.BASE_IMAGE,
            host_workdir.as_posix(),
            self.OUTPUT_IMAGE_NAME,
            extra_packages=[extra_package.as_posix()],
        )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=host_workdir.as_posix(),
                python_module=None,
            ),
            utils.DEFAULT_WORKDIR,
            utils.DEFAULT_HOME,
            requirements_path=None,
            setup_path=None,
            extra_requirements=None,
            extra_packages=[self.EXTRA_PACKAGE],
            extra_dirs=None,
            exposed_ports=None,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_mock.assert_called_once_with(
            [
                "docker",
                "build",
                "--no-cache",
                "-t",
                self.OUTPUT_IMAGE_NAME,
                "--rm",
                "-f-",
                host_workdir.as_posix(),
            ],
            input_str=make_dockerfile_mock.return_value,
        )
        assert image.name == self.OUTPUT_IMAGE_NAME
        assert image.default_home == self.HOME
        assert image.default_workdir == self.WORKDIR

    def test_build_image_not_found_extra_packages(self, make_dockerfile_mock):
        extra_package = f"./another_src/{self.EXTRA_PACKAGE}"
        expected_message = f'The extra_packages "{extra_package}" must exist.'

        with pytest.raises(ValueError) as exception:
            _ = build.build_image(
                self.BASE_IMAGE,
                self.HOST_WORKDIR,
                self.OUTPUT_IMAGE_NAME,
                extra_packages=[extra_package],
            )

        assert not make_dockerfile_mock.called
        assert str(exception.value) == expected_message

    def test_build_image_invalid_extra_packages(
        self, tmp_path, make_dockerfile_mock, execute_command_mock
    ):
        source_dir = tmp_path / self.SOURCE_DIR
        source_dir.mkdir()
        host_workdir = tmp_path / self.HOST_WORKDIR
        host_workdir.mkdir()
        another_dir = tmp_path / "another_dir"
        another_dir.mkdir()
        extra_package = another_dir / self.EXTRA_PACKAGE
        extra_package.write_text("")
        expected_message = (
            f'The extra_packages "{extra_package}" must be in "{host_workdir}".'
        )

        with pytest.raises(ValueError) as exception:
            _ = build.build_image(
                self.BASE_IMAGE,
                host_workdir.as_posix(),
                self.OUTPUT_IMAGE_NAME,
                extra_packages=[extra_package.as_posix()],
            )

        assert not make_dockerfile_mock.called
        assert str(exception.value) == expected_message

    def test_build_image_with_container_workdir(
        self, make_dockerfile_mock, execute_command_mock
    ):
        container_workdir = "/custom_workdir"

        image = build.build_image(
            self.BASE_IMAGE,
            self.HOST_WORKDIR,
            self.OUTPUT_IMAGE_NAME,
            container_workdir=container_workdir,
        )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=self.HOST_WORKDIR,
                python_module=None,
            ),
            container_workdir,
            utils.DEFAULT_HOME,
            requirements_path=None,
            setup_path=None,
            extra_requirements=None,
            extra_packages=None,
            extra_dirs=None,
            exposed_ports=None,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_mock.assert_called_once_with(
            [
                "docker",
                "build",
                "--no-cache",
                "-t",
                self.OUTPUT_IMAGE_NAME,
                "--rm",
                "-f-",
                self.HOST_WORKDIR,
            ],
            input_str=make_dockerfile_mock.return_value,
        )
        assert image.name == self.OUTPUT_IMAGE_NAME
        assert image.default_home == self.HOME
        assert image.default_workdir == container_workdir

    def test_build_image_with_container_home(
        self, make_dockerfile_mock, execute_command_mock
    ):
        container_home = "/custom_home"

        image = build.build_image(
            self.BASE_IMAGE,
            self.HOST_WORKDIR,
            self.OUTPUT_IMAGE_NAME,
            container_home=container_home,
        )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=self.HOST_WORKDIR,
                python_module=None,
            ),
            utils.DEFAULT_WORKDIR,
            container_home,
            requirements_path=None,
            setup_path=None,
            extra_requirements=None,
            extra_packages=None,
            extra_dirs=None,
            exposed_ports=None,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_mock.assert_called_once_with(
            [
                "docker",
                "build",
                "--no-cache",
                "-t",
                self.OUTPUT_IMAGE_NAME,
                "--rm",
                "-f-",
                self.HOST_WORKDIR,
            ],
            input_str=make_dockerfile_mock.return_value,
        )
        assert image.name == self.OUTPUT_IMAGE_NAME
        assert image.default_home == container_home
        assert image.default_workdir == self.WORKDIR

    def test_build_image_with_extra_dirs(
        self, make_dockerfile_mock, execute_command_mock
    ):
        extra_dirs = ["./another_src"]
        image = build.build_image(
            self.BASE_IMAGE,
            self.HOST_WORKDIR,
            self.OUTPUT_IMAGE_NAME,
            extra_dirs=extra_dirs,
        )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=self.HOST_WORKDIR,
                python_module=None,
            ),
            utils.DEFAULT_WORKDIR,
            utils.DEFAULT_HOME,
            requirements_path=None,
            setup_path=None,
            extra_requirements=None,
            extra_packages=None,
            extra_dirs=extra_dirs,
            exposed_ports=None,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_mock.assert_called_once_with(
            [
                "docker",
                "build",
                "--no-cache",
                "-t",
                self.OUTPUT_IMAGE_NAME,
                "--rm",
                "-f-",
                self.HOST_WORKDIR,
            ],
            input_str=make_dockerfile_mock.return_value,
        )
        assert image.name == self.OUTPUT_IMAGE_NAME
        assert image.default_home == self.HOME
        assert image.default_workdir == self.WORKDIR

    def test_build_image_with_exposed_ports(
        self, make_dockerfile_mock, execute_command_mock
    ):
        exposed_ports = [8080]
        image = build.build_image(
            self.BASE_IMAGE,
            self.HOST_WORKDIR,
            self.OUTPUT_IMAGE_NAME,
            exposed_ports=exposed_ports,
        )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=self.HOST_WORKDIR,
                python_module=None,
            ),
            utils.DEFAULT_WORKDIR,
            utils.DEFAULT_HOME,
            requirements_path=None,
            setup_path=None,
            extra_requirements=None,
            extra_packages=None,
            extra_dirs=None,
            exposed_ports=exposed_ports,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_mock.assert_called_once_with(
            [
                "docker",
                "build",
                "--no-cache",
                "-t",
                self.OUTPUT_IMAGE_NAME,
                "--rm",
                "-f-",
                self.HOST_WORKDIR,
            ],
            input_str=make_dockerfile_mock.return_value,
        )
        assert image.name == self.OUTPUT_IMAGE_NAME
        assert image.default_home == self.HOME
        assert image.default_workdir == self.WORKDIR

    def test_build_image_fail(
        self, make_dockerfile_mock, execute_command_return_code_1_mock
    ):
        command = [
            "docker",
            "build",
            "--no-cache",
            "-t",
            self.OUTPUT_IMAGE_NAME,
            "--rm",
            "-f-",
            self.HOST_WORKDIR,
        ]
        return_code = 1
        expected_message = textwrap.dedent(
            """
            Docker failed with error code {return_code}.
            Command: {command}
            """.format(
                return_code=return_code, command=" ".join(command)
            )
        )

        with pytest.raises(errors.DockerError) as exception:
            _ = build.build_image(
                self.BASE_IMAGE,
                self.HOST_WORKDIR,
                self.OUTPUT_IMAGE_NAME,
            )

        make_dockerfile_mock.assert_called_once_with(
            self.BASE_IMAGE,
            utils.Package(
                script=None,
                package_path=self.HOST_WORKDIR,
                python_module=None,
            ),
            utils.DEFAULT_WORKDIR,
            utils.DEFAULT_HOME,
            requirements_path=None,
            setup_path=None,
            extra_requirements=None,
            extra_packages=None,
            extra_dirs=None,
            exposed_ports=None,
            pip_command=self.PIP,
            python_command=self.PYTHON,
        )
        execute_command_return_code_1_mock.assert_called_once_with(
            command,
            input_str=make_dockerfile_mock.return_value,
        )
        assert exception.value.message == expected_message
        assert exception.value.cmd == command
        assert exception.value.exit_code == return_code


class TestLocalUtil:
    @mock.patch("subprocess.Popen", MockedPopen)
    def test_execute_command(self, textiowrapper_mock):
        return_code = local_util.execute_command(["ls"])

        assert return_code == 0

    @mock.patch("subprocess.Popen", MockedPopen)
    def test_execute_command_with_input_str(self, textiowrapper_mock):
        return_code = local_util.execute_command(["ls"], "Test string.")

        assert return_code == 0


class TestErrors:
    def test_raise_docker_error_with_command(self):
        command = ["ls", "-l"]
        return_code = 1
        expected_message = textwrap.dedent(
            """
            Docker failed with error code {return_code}.
            Command: {command}
            """.format(
                return_code=return_code, command=" ".join(command)
            )
        )

        with pytest.raises(errors.DockerError) as exception:
            errors.raise_docker_error_with_command(command, return_code)

        assert exception.value.message == expected_message
        assert exception.value.cmd == command
        assert exception.value.exit_code == return_code


class TestUtils:
    IMAGE_URI = "test_image:latest"

    def test_check_image_exists_locally(self, docker_client_mock):
        assert utils.check_image_exists_locally(self.IMAGE_URI) is True

    def test_check_image_exists_locally_image_not_found(
        self, docker_client_mock_image_get_not_found
    ):
        assert utils.check_image_exists_locally(self.IMAGE_URI) is False

    def test_check_image_exists_locally_image_api_error(
        self, docker_client_mock_image_get_api_error
    ):
        assert utils.check_image_exists_locally(self.IMAGE_URI) is False



# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.protobuf import struct_pb2  # type: ignore
from google.protobuf import timestamp_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1beta1",
    manifest={
        "ModelEvaluationSlice",
    },
)


class ModelEvaluationSlice(proto.Message):
    r"""A collection of metrics calculated by comparing Model's
    predictions on a slice of the test data against ground truth
    annotations.

    Attributes:
        name (str):
            Output only. The resource name of the
            ModelEvaluationSlice.
        slice_ (google.cloud.aiplatform_v1beta1.types.ModelEvaluationSlice.Slice):
            Output only. The slice of the test data that
            is used to evaluate the Model.
        metrics_schema_uri (str):
            Output only. Points to a YAML file stored on Google Cloud
            Storage describing the
            [metrics][google.cloud.aiplatform.v1beta1.ModelEvaluationSlice.metrics]
            of this ModelEvaluationSlice. The schema is defined as an
            OpenAPI 3.0.2 `Schema
            Object <https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.2.md#schemaObject>`__.
        metrics (google.protobuf.struct_pb2.Value):
            Output only. Sliced evaluation metrics of the Model. The
            schema of the metrics is stored in
            [metrics_schema_uri][google.cloud.aiplatform.v1beta1.ModelEvaluationSlice.metrics_schema_uri]
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Timestamp when this
            ModelEvaluationSlice was created.
    """

    class Slice(proto.Message):
        r"""Definition of a slice.

        Attributes:
            dimension (str):
                Output only. The dimension of the slice. Well-known
                dimensions are:

                -  ``annotationSpec``: This slice is on the test data that
                   has either ground truth or prediction with
                   [AnnotationSpec.display_name][google.cloud.aiplatform.v1beta1.AnnotationSpec.display_name]
                   equals to
                   [value][google.cloud.aiplatform.v1beta1.ModelEvaluationSlice.Slice.value].
            value (str):
                Output only. The value of the dimension in
                this slice.
        """

        dimension: str = proto.Field(
            proto.STRING,
            number=1,
        )
        value: str = proto.Field(
            proto.STRING,
            number=2,
        )

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    slice_: Slice = proto.Field(
        proto.MESSAGE,
        number=2,
        message=Slice,
    )
    metrics_schema_uri: str = proto.Field(
        proto.STRING,
        number=3,
    )
    metrics: struct_pb2.Value = proto.Field(
        proto.MESSAGE,
        number=4,
        message=struct_pb2.Value,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=5,
        message=timestamp_pb2.Timestamp,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import MutableMapping, MutableSequence

import proto  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1.schema.trainingjob.definition",
    manifest={
        "AutoMlVideoClassification",
        "AutoMlVideoClassificationInputs",
    },
)


class AutoMlVideoClassification(proto.Message):
    r"""A TrainingJob that trains and uploads an AutoML Video
    Classification Model.

    Attributes:
        inputs (google.cloud.aiplatform.v1.schema.trainingjob.definition_v1.types.AutoMlVideoClassificationInputs):
            The input parameters of this TrainingJob.
    """

    inputs: "AutoMlVideoClassificationInputs" = proto.Field(
        proto.MESSAGE,
        number=1,
        message="AutoMlVideoClassificationInputs",
    )


class AutoMlVideoClassificationInputs(proto.Message):
    r"""

    Attributes:
        model_type (google.cloud.aiplatform.v1.schema.trainingjob.definition_v1.types.AutoMlVideoClassificationInputs.ModelType):

    """

    class ModelType(proto.Enum):
        r"""

        Values:
            MODEL_TYPE_UNSPECIFIED (0):
                Should not be set.
            CLOUD (1):
                A model best tailored to be used within
                Google Cloud, and which cannot be exported.
                Default.
            MOBILE_VERSATILE_1 (2):
                A model that, in addition to being available
                within Google Cloud, can also be exported (see
                ModelService.ExportModel) as a TensorFlow or
                TensorFlow Lite model and used on a mobile or
                edge device afterwards.
            MOBILE_JETSON_VERSATILE_1 (3):
                A model that, in addition to being available
                within Google Cloud, can also be exported (see
                ModelService.ExportModel) to a Jetson device
                afterwards.
        """
        MODEL_TYPE_UNSPECIFIED = 0
        CLOUD = 1
        MOBILE_VERSATILE_1 = 2
        MOBILE_JETSON_VERSATILE_1 = 3

    model_type: ModelType = proto.Field(
        proto.ENUM,
        number=1,
        enum=ModelType,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Dict, List, Optional, Sequence, Tuple

from google.auth import credentials as auth_credentials
from google.protobuf import field_mask_pb2
from google.cloud.aiplatform import base
from google.cloud.aiplatform.compat.types import (
    matching_engine_deployed_index_ref as gca_matching_engine_deployed_index_ref,
    matching_engine_index as gca_matching_engine_index,
)
from google.cloud.aiplatform import initializer
from google.cloud.aiplatform.matching_engine import (
    matching_engine_index_config,
)
from google.cloud.aiplatform import utils

_LOGGER = base.Logger(__name__)


class MatchingEngineIndex(base.VertexAiResourceNounWithFutureManager):
    """Matching Engine index resource for Vertex AI."""

    client_class = utils.IndexClientWithOverride

    _resource_noun = "indexes"
    _getter_method = "get_index"
    _list_method = "list_indexes"
    _delete_method = "delete_index"
    _parse_resource_name_method = "parse_index_path"
    _format_resource_name_method = "index_path"

    def __init__(
        self,
        index_name: str,
        project: Optional[str] = None,
        location: Optional[str] = None,
        credentials: Optional[auth_credentials.Credentials] = None,
    ):
        """Retrieves an existing index given an index name or ID.

        Example Usage:

            my_index = aiplatform.MatchingEngineIndex(
                index_name='projects/123/locations/us-central1/indexes/my_index_id'
            )
            or
            my_index = aiplatform.MatchingEngineIndex(
                index_name='my_index_id'
            )

        Args:
            index_name (str):
                Required. A fully-qualified index resource name or a index ID.
                Example: "projects/123/locations/us-central1/indexes/my_index_id"
                or "my_index_id" when project and location are initialized or passed.
            project (str):
                Optional. Project to retrieve index from. If not set, project
                set in aiplatform.init will be used.
            location (str):
                Optional. Location to retrieve index from. If not set, location
                set in aiplatform.init will be used.
            credentials (auth_credentials.Credentials):
                Optional. Custom credentials to use to retrieve this Index. Overrides
                credentials set in aiplatform.init.
        """

        super().__init__(
            project=project,
            location=location,
            credentials=credentials,
            resource_name=index_name,
        )
        self._gca_resource = self._get_gca_resource(resource_name=index_name)

    @property
    def description(self) -> str:
        """Description of the index."""
        self._assert_gca_resource_is_available()
        return self._gca_resource.description

    @classmethod
    @base.optional_sync()
    def _create(
        cls,
        display_name: str,
        contents_delta_uri: str,
        config: matching_engine_index_config.MatchingEngineIndexConfig,
        description: Optional[str] = None,
        labels: Optional[Dict[str, str]] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        credentials: Optional[auth_credentials.Credentials] = None,
        request_metadata: Optional[Sequence[Tuple[str, str]]] = (),
        sync: bool = True,
    ) -> "MatchingEngineIndex":
        """Creates a MatchingEngineIndex resource.

        Args:
            display_name (str):
                Required. The display name of the Index.
                The name can be up to 128 characters long and
                can be consist of any UTF-8 characters.
            contents_delta_uri (str):
                Required. Allows inserting, updating  or deleting the contents of the Matching Engine Index.
                The string must be a valid Google Cloud Storage directory path. If this
                field is set when calling IndexService.UpdateIndex, then no other
                Index field can be  also updated as part of the same call.
                The expected structure and format of the files this URI points to is
                described at
                https://docs.google.com/document/d/12DLVB6Nq6rdv8grxfBsPhUA283KWrQ9ZenPBp0zUC30
            config (matching_engine_index_config.MatchingEngineIndexConfig):
                Required. The configuration with regard to the algorithms used for efficient search.
            description (str):
                Optional. The description of the Index.
            labels (Dict[str, str]):
                Optional. The labels with user-defined
                metadata to organize your Index.
                Label keys and values can be no longer than 64
                characters (Unicode codepoints), can only
                contain lowercase letters, numeric characters,
                underscores and dashes. International characters
                are allowed.
                See https://goo.gl/xmQnxf for more information
                on and examples of labels. No more than 64 user
                labels can be associated with one
                Index(System labels are excluded)."
                System reserved label keys are prefixed with
                "aiplatform.googleapis.com/" and are immutable.
            project (str):
                Optional. Project to create EntityType in. If not set, project
                set in aiplatform.init will be used.
            location (str):
                Optional. Location to create EntityType in. If not set, location
                set in aiplatform.init will be used.
            credentials (auth_credentials.Credentials):
                Optional. Custom credentials to use to create EntityTypes. Overrides
                credentials set in aiplatform.init.
            request_metadata (Sequence[Tuple[str, str]]):
                Optional. Strings which should be sent along with the request as metadata.
            encryption_spec (str):
                Optional. Customer-managed encryption key
                spec for data storage. If set, both of the
                online and offline data storage will be secured
                by this key.
            sync (bool):
                Optional. Whether to execute this creation synchronously. If False, this method
                will be executed in concurrent Future and any downstream object will
                be immediately returned and synced when the Future has completed.

        Returns:
            MatchingEngineIndex - Index resource object

        """
        gapic_index = gca_matching_engine_index.Index(
            display_name=display_name,
            description=description,
            metadata={
                "config": config.as_dict(),
                "contentsDeltaUri": contents_delta_uri,
            },
        )

        if labels:
            utils.validate_labels(labels)
            gapic_index.labels = labels

        api_client = cls._instantiate_client(location=location, credentials=credentials)

        create_lro = api_client.create_index(
            parent=initializer.global_config.common_location_path(
                project=project, location=location
            ),
            index=gapic_index,
            metadata=request_metadata,
        )

        _LOGGER.log_create_with_lro(cls, create_lro)

        created_index = create_lro.result(timeout=None)

        _LOGGER.log_create_complete(cls, created_index, "index")

        index_obj = cls(
            index_name=created_index.name,
            project=project,
            location=location,
            credentials=credentials,
        )

        return index_obj

    def update_metadata(
        self,
        display_name: Optional[str] = None,
        description: Optional[str] = None,
        labels: Optional[Dict[str, str]] = None,
        request_metadata: Optional[Sequence[Tuple[str, str]]] = (),
    ) -> "MatchingEngineIndex":
        """Updates the metadata for this index.

        Args:
            display_name (str):
                Optional. The display name of the Index.
                The name can be up to 128 characters long and
                can be consist of any UTF-8 characters.
            description (str):
                Optional. The description of the Index.
            labels (Dict[str, str]):
                Optional. The labels with user-defined
                metadata to organize your Indexs.
                Label keys and values can be no longer than 64
                characters (Unicode codepoints), can only
                contain lowercase letters, numeric characters,
                underscores and dashes. International characters
                are allowed.
                See https://goo.gl/xmQnxf for more information
                on and examples of labels. No more than 64 user
                labels can be associated with one Index
                (System labels are excluded)."
                System reserved label keys are prefixed with
                "aiplatform.googleapis.com/" and are immutable.
            request_metadata (Sequence[Tuple[str, str]]):
                Optional. Strings which should be sent along with the request as metadata.

        Returns:
            MatchingEngineIndex - The updated index resource object.
        """

        self.wait()

        update_mask = list()

        if labels:
            utils.validate_labels(labels)
            update_mask.append("labels")

        if display_name is not None:
            update_mask.append("display_name")

        if description is not None:
            update_mask.append("description")

        update_mask = field_mask_pb2.FieldMask(paths=update_mask)

        gapic_index = gca_matching_engine_index.Index(
            name=self.resource_name,
            display_name=display_name,
            description=description,
            labels=labels,
        )

        _LOGGER.log_action_start_against_resource(
            "Updating",
            "index",
            self,
        )

        update_lro = self.api_client.update_index(
            index=gapic_index,
            update_mask=update_mask,
            metadata=request_metadata,
        )

        _LOGGER.log_action_started_against_resource_with_lro(
            "Update", "index", self.__class__, update_lro
        )

        self._gca_resource = update_lro.result()

        _LOGGER.log_action_completed_against_resource("index", "Updated", self)

        return self

    def update_embeddings(
        self,
        contents_delta_uri: str,
        is_complete_overwrite: Optional[bool] = None,
        request_metadata: Optional[Sequence[Tuple[str, str]]] = (),
    ) -> "MatchingEngineIndex":
        """Updates the embeddings for this index.

        Args:
            contents_delta_uri (str):
                Required. Allows inserting, updating  or deleting the contents of the Matching Engine Index.
                The string must be a valid Google Cloud Storage directory path. If this
                field is set when calling IndexService.UpdateIndex, then no other
                Index field can be  also updated as part of the same call.
                The expected structure and format of the files this URI points to is
                described at
                https://docs.google.com/document/d/12DLVB6Nq6rdv8grxfBsPhUA283KWrQ9ZenPBp0zUC30
            is_complete_overwrite (bool):
                Optional. If this field is set together with contentsDeltaUri when calling IndexService.UpdateIndex,
                then existing content of the Index will be replaced by the data from the contentsDeltaUri.
            request_metadata (Sequence[Tuple[str, str]]):
                Optional. Strings which should be sent along with the request as metadata.

        Returns:
            MatchingEngineIndex - The updated index resource object.
        """

        self.wait()

        update_mask = list()

        if contents_delta_uri or is_complete_overwrite:
            update_mask.append("metadata")

        update_mask = field_mask_pb2.FieldMask(paths=update_mask)

        gapic_index = gca_matching_engine_index.Index(
            name=self.resource_name,
            metadata={
                "contentsDeltaUri": contents_delta_uri,
                "isCompleteOverwrite": is_complete_overwrite,
            },
        )

        _LOGGER.log_action_start_against_resource(
            "Updating",
            "index",
            self,
        )

        update_lro = self.api_client.update_index(
            index=gapic_index,
            update_mask=update_mask,
            metadata=request_metadata,
        )

        _LOGGER.log_action_started_against_resource_with_lro(
            "Update", "index", self.__class__, update_lro
        )

        self._gca_resource = update_lro.result(timeout=None)

        _LOGGER.log_action_completed_against_resource("index", "Updated", self)

        return self

    @property
    def deployed_indexes(
        self,
    ) -> List[gca_matching_engine_deployed_index_ref.DeployedIndexRef]:
        """Returns a list of deployed index references that originate from this index.

        Returns:
            List[gca_matching_engine_deployed_index_ref.DeployedIndexRef] - Deployed index references
        """

        self.wait()

        return self._gca_resource.deployed_indexes

    @classmethod
    def create_tree_ah_index(
        cls,
        display_name: str,
        contents_delta_uri: str,
        dimensions: int,
        approximate_neighbors_count: int,
        leaf_node_embedding_count: Optional[int] = None,
        leaf_nodes_to_search_percent: Optional[float] = None,
        distance_measure_type: Optional[
            matching_engine_index_config.DistanceMeasureType
        ] = None,
        description: Optional[str] = None,
        labels: Optional[Dict[str, str]] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        credentials: Optional[auth_credentials.Credentials] = None,
        request_metadata: Optional[Sequence[Tuple[str, str]]] = (),
        sync: bool = True,
    ) -> "MatchingEngineIndex":
        """Creates a MatchingEngineIndex resource that uses the tree-AH algorithm.

        Example Usage:

            my_index = aiplatform.Index.create_tree_ah_index(
                display_name="my_display_name",
                contents_delta_uri="gs://my_bucket/embeddings",
                dimensions=1,
                approximate_neighbors_count=150,
                distance_measure_type="SQUARED_L2_DISTANCE",
                leaf_node_embedding_count=100,
                leaf_nodes_to_search_percent=50,
                description="my description",
                labels={ "label_name": "label_value" },
            )

        Args:
            display_name (str):
                Required. The display name of the Index.
                The name can be up to 128 characters long and
                can be consist of any UTF-8 characters.
            contents_delta_uri (str):
                Required. Allows inserting, updating  or deleting the contents of the Matching Engine Index.
                The string must be a valid Google Cloud Storage directory path. If this
                field is set when calling IndexService.UpdateIndex, then no other
                Index field can be  also updated as part of the same call.
                The expected structure and format of the files this URI points to is
                described at
                https://docs.google.com/document/d/12DLVB6Nq6rdv8grxfBsPhUA283KWrQ9ZenPBp0zUC30
            dimensions (int):
                Required. The number of dimensions of the input vectors.
            approximate_neighbors_count (int):
                Required. The default number of neighbors to find via approximate search before exact reordering is
                performed. Exact reordering is a procedure where results returned by an
                approximate search algorithm are reordered via a more expensive distance computation.
            leaf_node_embedding_count (int):
                Optional. Number of embeddings on each leaf node. The default value is 1000 if not set.
            leaf_nodes_to_search_percent (float):
                Optional. The default percentage of leaf nodes that any query may be searched. Must be in
                range 1-100, inclusive. The default value is 10 (means 10%) if not set.
            distance_measure_type (matching_engine_index_config.DistanceMeasureType):
                Optional. The distance measure used in nearest neighbor search.
            description (str):
                Optional. The description of the Index.
            labels (Dict[str, str]):
                Optional. The labels with user-defined
                metadata to organize your Index.
                Label keys and values can be no longer than 64
                characters (Unicode codepoints), can only
                contain lowercase letters, numeric characters,
                underscores and dashes. International characters
                are allowed.
                See https://goo.gl/xmQnxf for more information
                on and examples of labels. No more than 64 user
                labels can be associated with one
                Index(System labels are excluded)."
                System reserved label keys are prefixed with
                "aiplatform.googleapis.com/" and are immutable.
            project (str):
                Optional. Project to create EntityType in. If not set, project
                set in aiplatform.init will be used.
            location (str):
                Optional. Location to create EntityType in. If not set, location
                set in aiplatform.init will be used.
            credentials (auth_credentials.Credentials):
                Optional. Custom credentials to use to create EntityTypes. Overrides
                credentials set in aiplatform.init.
            request_metadata (Sequence[Tuple[str, str]]):
                Optional. Strings which should be sent along with the request as metadata.
            encryption_spec (str):
                Optional. Customer-managed encryption key
                spec for data storage. If set, both of the
                online and offline data storage will be secured
                by this key.
            sync (bool):
                Optional. Whether to execute this creation synchronously. If False, this method
                will be executed in concurrent Future and any downstream object will
                be immediately returned and synced when the Future has completed.

        Returns:
            MatchingEngineIndex - Index resource object

        """

        algorithm_config = matching_engine_index_config.TreeAhConfig(
            leaf_node_embedding_count=leaf_node_embedding_count,
            leaf_nodes_to_search_percent=leaf_nodes_to_search_percent,
        )

        config = matching_engine_index_config.MatchingEngineIndexConfig(
            dimensions=dimensions,
            algorithm_config=algorithm_config,
            approximate_neighbors_count=approximate_neighbors_count,
            distance_measure_type=distance_measure_type,
        )

        return cls._create(
            display_name=display_name,
            contents_delta_uri=contents_delta_uri,
            config=config,
            description=description,
            labels=labels,
            project=project,
            location=location,
            credentials=credentials,
            request_metadata=request_metadata,
            sync=sync,
        )

    @classmethod
    def create_brute_force_index(
        cls,
        display_name: str,
        contents_delta_uri: str,
        dimensions: int,
        distance_measure_type: Optional[
            matching_engine_index_config.DistanceMeasureType
        ] = None,
        description: Optional[str] = None,
        labels: Optional[Dict[str, str]] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        credentials: Optional[auth_credentials.Credentials] = None,
        request_metadata: Optional[Sequence[Tuple[str, str]]] = (),
        sync: bool = True,
    ) -> "MatchingEngineIndex":
        """Creates a MatchingEngineIndex resource that uses the brute force algorithm.

        Example Usage:

            my_index = aiplatform.Index.create_brute_force_index(
                display_name="my_display_name",
                contents_delta_uri="gs://my_bucket/embeddings",
                dimensions=1,
                approximate_neighbors_count=150,
                distance_measure_type="SQUARED_L2_DISTANCE",
                description="my description",
                labels={ "label_name": "label_value" },
            )

        Args:
            display_name (str):
                Required. The display name of the Index.
                The name can be up to 128 characters long and
                can be consist of any UTF-8 characters.
            contents_delta_uri (str):
                Required. Allows inserting, updating  or deleting the contents of the Matching Engine Index.
                The string must be a valid Google Cloud Storage directory path. If this
                field is set when calling IndexService.UpdateIndex, then no other
                Index field can be  also updated as part of the same call.
                The expected structure and format of the files this URI points to is
                described at
                https://docs.google.com/document/d/12DLVB6Nq6rdv8grxfBsPhUA283KWrQ9ZenPBp0zUC30
            dimensions (int):
                Required. The number of dimensions of the input vectors.
            distance_measure_type (matching_engine_index_config.DistanceMeasureType):
                Optional. The distance measure used in nearest neighbor search.
            description (str):
                Optional. The description of the Index.
            labels (Dict[str, str]):
                Optional. The labels with user-defined
                metadata to organize your Index.
                Label keys and values can be no longer than 64
                characters (Unicode codepoints), can only
                contain lowercase letters, numeric characters,
                underscores and dashes. International characters
                are allowed.
                See https://goo.gl/xmQnxf for more information
                on and examples of labels. No more than 64 user
                labels can be associated with one
                Index(System labels are excluded)."
                System reserved label keys are prefixed with
                "aiplatform.googleapis.com/" and are immutable.
            project (str):
                Optional. Project to create EntityType in. If not set, project
                set in aiplatform.init will be used.
            location (str):
                Optional. Location to create EntityType in. If not set, location
                set in aiplatform.init will be used.
            credentials (auth_credentials.Credentials):
                Optional. Custom credentials to use to create EntityTypes. Overrides
                credentials set in aiplatform.init.
            request_metadata (Sequence[Tuple[str, str]]):
                Optional. Strings which should be sent along with the request as metadata.
            encryption_spec (str):
                Optional. Customer-managed encryption key
                spec for data storage. If set, both of the
                online and offline data storage will be secured
                by this key.
            sync (bool):
                Optional. Whether to execute this creation synchronously. If False, this method
                will be executed in concurrent Future and any downstream object will
                be immediately returned and synced when the Future has completed.

        Returns:
            MatchingEngineIndex - Index resource object

        """

        algorithm_config = matching_engine_index_config.BruteForceConfig()

        config = matching_engine_index_config.MatchingEngineIndexConfig(
            dimensions=dimensions,
            algorithm_config=algorithm_config,
            distance_measure_type=distance_measure_type,
        )

        return cls._create(
            display_name=display_name,
            contents_delta_uri=contents_delta_uri,
            config=config,
            description=description,
            labels=labels,
            project=project,
            location=location,
            credentials=credentials,
            request_metadata=request_metadata,
            sync=sync,
        )



# -*- coding: utf-8 -*-

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import joblib
import logging
import os
import pickle

import numpy as np
import xgboost as xgb

from google.cloud.aiplatform.constants import prediction
from google.cloud.aiplatform.utils import prediction_utils
from google.cloud.aiplatform.prediction.predictor import Predictor


class XgboostPredictor(Predictor):
    """Default Predictor implementation for Xgboost models."""

    def __init__(self):
        return

    def load(self, artifacts_uri: str) -> None:
        """Loads the model artifact.

        Args:
            artifacts_uri (str):
                Required. The value of the environment variable AIP_STORAGE_URI.

        Raises:
            ValueError: If there's no required model files provided in the artifacts
                uri.
        """
        prediction_utils.download_model_artifacts(artifacts_uri)
        if os.path.exists(prediction.MODEL_FILENAME_BST):
            booster = xgb.Booster(model_file=prediction.MODEL_FILENAME_BST)
        elif os.path.exists(prediction.MODEL_FILENAME_JOBLIB):
            try:
                booster = joblib.load(prediction.MODEL_FILENAME_JOBLIB)
            except KeyError:
                logging.info(
                    "Loading model using joblib failed. "
                    "Loading model using xgboost.Booster instead."
                )
                booster = xgb.Booster()
                booster.load_model(prediction.MODEL_FILENAME_JOBLIB)
        elif os.path.exists(prediction.MODEL_FILENAME_PKL):
            booster = pickle.load(open(prediction.MODEL_FILENAME_PKL, "rb"))
        else:
            valid_filenames = [
                prediction.MODEL_FILENAME_BST,
                prediction.MODEL_FILENAME_JOBLIB,
                prediction.MODEL_FILENAME_PKL,
            ]
            raise ValueError(
                f"One of the following model files must be provided: {valid_filenames}."
            )
        self._booster = booster

    def preprocess(self, prediction_input: dict) -> xgb.DMatrix:
        """Converts the request body to a Data Matrix before prediction.
        Args:
            prediction_input (dict):
                Required. The prediction input that needs to be preprocessed.
        Returns:
            The preprocessed prediction input.
        """
        instances = prediction_input["instances"]
        return xgb.DMatrix(instances)

    def predict(self, instances: xgb.DMatrix) -> np.ndarray:
        """Performs prediction.

        Args:
            instances (xgb.DMatrix):
                Required. The instance(s) used for performing prediction.

        Returns:
            Prediction results.
        """
        return self._booster.predict(instances)

    def postprocess(self, prediction_results: np.ndarray) -> dict:
        """Converts numpy array to a dict.
        Args:
            prediction_results (np.ndarray):
                Required. The prediction results.
        Returns:
            The postprocessed prediction results.
        """
        return {"predictions": prediction_results.tolist()}



# -*- coding: utf-8 -*-
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Use this file to store global variables that will be shared across multiple tests

import dataclasses
from datetime import datetime

from google.protobuf import timestamp_pb2, duration_pb2

from google.cloud.aiplatform.utils import source_utils
from google.cloud.aiplatform import explain
from google.cloud.aiplatform import utils

from google.cloud.aiplatform.compat.services import (
    model_service_client,
)

from google.cloud.aiplatform.compat.types import (
    custom_job,
    encryption_spec,
    endpoint,
    io,
    model,
    tensorboard_data,
    tensorboard_experiment,
    tensorboard_run,
    tensorboard_time_series,
)


@dataclasses.dataclass(frozen=True)
class ProjectConstants:
    """Defines project-specific constants used by tests."""

    _TEST_PROJECT = "test-project"
    _TEST_LOCATION = "us-central1"
    _TEST_ENCRYPTION_KEY_NAME = "key_1234"
    _TEST_ENCRYPTION_SPEC = encryption_spec.EncryptionSpec(
        kms_key_name=_TEST_ENCRYPTION_KEY_NAME
    )
    _TEST_PARENT = f"projects/{_TEST_PROJECT}/locations/{_TEST_LOCATION}"
    _TEST_SERVICE_ACCOUNT = "vinnys@my-project.iam.gserviceaccount.com"
    _TEST_LABELS = {"my_key": "my_value"}


@dataclasses.dataclass(frozen=True)
class TrainingJobConstants:
    """Defines constants used by tests that create training jobs."""

    _TEST_OUTPUT_PYTHON_PACKAGE_PATH = "gs://test-staging-bucket/trainer.tar.gz"
    _TEST_MODULE_NAME = (
        f"{source_utils._TrainingScriptPythonPackager._ROOT_MODULE}.task"
    )
    _TEST_LOCAL_SCRIPT_FILE_NAME = "____test____script.py"
    _TEST_REQUIREMENTS = ["pandas", "numpy", "tensorflow"]
    _TEST_ENVIRONMENT_VARIABLES = {
        "MY_PATH": "/path/to/my_path",
    }
    _TEST_REPLICA_COUNT = 1
    _TEST_MACHINE_TYPE = "n1-standard-4"
    _TEST_ACCELERATOR_TYPE = "NVIDIA_TESLA_K80"
    _TEST_ACCELERATOR_COUNT = 1
    _TEST_BOOT_DISK_TYPE = "pd-standard"
    _TEST_BOOT_DISK_SIZE_GB = 300
    _TEST_REDUCTION_SERVER_REPLICA_COUNT = 1
    _TEST_REDUCTION_SERVER_MACHINE_TYPE = "n1-highcpu-16"
    _TEST_REDUCTION_SERVER_CONTAINER_URI = (
        "us-docker.pkg.dev/vertex-ai-restricted/training/reductionserver:latest"
    )
    _TEST_STAGING_BUCKET = "gs://test-staging-bucket"
    _TEST_DISPLAY_NAME = "my_job_1234"
    _TEST_BASE_OUTPUT_DIR = f"{_TEST_STAGING_BUCKET}/{_TEST_DISPLAY_NAME}"
    _TEST_ENABLE_WEB_ACCESS = True
    _TEST_WEB_ACCESS_URIS = {"workerpool0-0": "uri"}
    _TEST_TRAINING_CONTAINER_IMAGE = "gcr.io/test-training/container:image"

    _TEST_RUN_ARGS = ["-v", "0.1", "--test=arg"]

    _TEST_WORKER_POOL_SPEC = [
        {
            "machine_spec": {
                "machine_type": "n1-standard-4",
                "accelerator_type": "NVIDIA_TESLA_K80",
                "accelerator_count": 1,
            },
            "replica_count": 1,
            "disk_spec": {"boot_disk_type": "pd-ssd", "boot_disk_size_gb": 100},
            "container_spec": {
                "image_uri": _TEST_TRAINING_CONTAINER_IMAGE,
                "command": [],
                "args": _TEST_RUN_ARGS,
            },
        }
    ]
    _TEST_ID = "1028944691210842416"
    _TEST_NETWORK = (
        f"projects/{ProjectConstants._TEST_PROJECT}/global/networks/{_TEST_ID}"
    )
    _TEST_TIMEOUT = 8000
    _TEST_RESTART_JOB_ON_WORKER_RESTART = True

    _TEST_BASE_CUSTOM_JOB_PROTO = custom_job.CustomJob(
        display_name=_TEST_DISPLAY_NAME,
        job_spec=custom_job.CustomJobSpec(
            worker_pool_specs=_TEST_WORKER_POOL_SPEC,
            base_output_directory=io.GcsDestination(
                output_uri_prefix=_TEST_BASE_OUTPUT_DIR
            ),
            scheduling=custom_job.Scheduling(
                timeout=duration_pb2.Duration(seconds=_TEST_TIMEOUT),
                restart_job_on_worker_restart=_TEST_RESTART_JOB_ON_WORKER_RESTART,
            ),
            service_account=ProjectConstants._TEST_SERVICE_ACCOUNT,
            network=_TEST_NETWORK,
        ),
        labels=ProjectConstants._TEST_LABELS,
        encryption_spec=ProjectConstants._TEST_ENCRYPTION_SPEC,
    )


@dataclasses.dataclass(frozen=True)
class ModelConstants:
    """Defines constants used by tests that create model resources."""

    _TEST_MODEL_NAME = "123"
    _TEST_ID = "1028944691210842416"
    _TEST_VERSION_ID = "2"
    _TEST_MODEL_RESOURCE_NAME = model_service_client.ModelServiceClient.model_path(
        ProjectConstants._TEST_PROJECT, ProjectConstants._TEST_LOCATION, _TEST_ID
    )
    _TEST_MODEL_PARENT = f"projects/{ProjectConstants._TEST_PROJECT}/locations/{ProjectConstants._TEST_LOCATION}/models/{_TEST_MODEL_NAME}"
    _TEST_VERSION_ALIAS_1 = "myalias"
    _TEST_VERSION_ALIAS_2 = "youralias"
    _TEST_MODEL_VERSION_DESCRIPTION_2 = "My version 2 description"
    _TEST_SERVING_CONTAINER_IMAGE = "gcr.io/test-serving/container:image"
    _TEST_EXPLANATION_PARAMETERS = explain.ExplanationParameters(
        {"sampled_shapley_attribution": {"path_count": 10}}
    )
    _TEST_LABEL = {"team": "experimentation", "trial_id": "x435"}
    _TEST_MODEL_OBJ_WITH_VERSION = model.Model(
        version_id=_TEST_VERSION_ID,
        create_time=timestamp_pb2.Timestamp(),
        update_time=timestamp_pb2.Timestamp(),
        display_name=_TEST_MODEL_NAME,
        name=f"{_TEST_MODEL_PARENT}@{_TEST_VERSION_ID}",
        version_aliases=[_TEST_VERSION_ALIAS_1, _TEST_VERSION_ALIAS_2],
        version_description=_TEST_MODEL_VERSION_DESCRIPTION_2,
    )


@dataclasses.dataclass(frozen=True)
class EndpointConstants:
    """Defines constants used by tests that create endpoints."""

    _TEST_DISPLAY_NAME = "test-display-name"
    _TEST_DISPLAY_NAME_2 = "test-display-name-2"
    _TEST_DISPLAY_NAME_3 = "test-display-name-3"
    _TEST_ID = "1028944691210842416"
    _TEST_ID_2 = "4366591682456584192"
    _TEST_ID_3 = "5820582938582924817"
    _TEST_ENDPOINT_NAME = f"projects/{ProjectConstants._TEST_PROJECT}/locations/{ProjectConstants._TEST_LOCATION}/endpoints/{_TEST_ID}"
    _TEST_DISPLAY_NAME = "test-display-name"
    _TEST_DEPLOYED_MODELS = [
        endpoint.DeployedModel(id=_TEST_ID, display_name=_TEST_DISPLAY_NAME),
        endpoint.DeployedModel(id=_TEST_ID_2, display_name=_TEST_DISPLAY_NAME_2),
        endpoint.DeployedModel(id=_TEST_ID_3, display_name=_TEST_DISPLAY_NAME_3),
    ]
    _TEST_TRAFFIC_SPLIT = {_TEST_ID: 0, _TEST_ID_2: 100, _TEST_ID_3: 0}


@dataclasses.dataclass(frozen=True)
class TensorboardConstants:
    """Defines constants used by tests that create Tensorboard resources."""

    _TEST_ID = "1028944691210842416"
    _TEST_DISPLAY_NAME = "my_tensorboard_1234"
    _TEST_TENSORBOARD_NAME = f"{ProjectConstants._TEST_PARENT}/tensorboards/{_TEST_ID}"
    _TEST_TENSORBOARD_EXPERIMENT_ID = "test-experiment"
    _TEST_TENSORBOARD_EXPERIMENT_NAME = (
        f"{_TEST_TENSORBOARD_NAME}/experiments/{_TEST_TENSORBOARD_EXPERIMENT_ID}"
    )

    _TEST_TENSORBOARD_RUN_ID = "test-run"
    _TEST_TENSORBOARD_RUN_NAME = (
        f"{_TEST_TENSORBOARD_EXPERIMENT_NAME}/runs/{_TEST_TENSORBOARD_RUN_ID}"
    )
    _TEST_TENSORBOARD_RUN = tensorboard_run.TensorboardRun(
        name=_TEST_TENSORBOARD_RUN_NAME,
        display_name=_TEST_DISPLAY_NAME,
    )
    _TEST_TENSORBOARD_TIME_SERIES_ID = "test-time-series"
    _TEST_TENSORBOARD_TIME_SERIES_NAME = (
        f"{_TEST_TENSORBOARD_RUN_NAME}/timeSeries/{_TEST_TENSORBOARD_TIME_SERIES_ID}"
    )
    _TEST_TIME_SERIES_DISPLAY_NAME = "accuracy"
    _TEST_TENSORBOARD_TIME_SERIES = tensorboard_time_series.TensorboardTimeSeries(
        name=_TEST_TENSORBOARD_TIME_SERIES_NAME,
        display_name=_TEST_TIME_SERIES_DISPLAY_NAME,
        value_type=tensorboard_time_series.TensorboardTimeSeries.ValueType.SCALAR,
    )
    _TEST_TENSORBOARD_TIME_SERIES_DATA = tensorboard_data.TimeSeriesData(
        tensorboard_time_series_id=_TEST_TENSORBOARD_TIME_SERIES_ID,
        value_type=tensorboard_time_series.TensorboardTimeSeries.ValueType.SCALAR,
        values=[
            tensorboard_data.TimeSeriesDataPoint(
                scalar=tensorboard_data.Scalar(value=1.0),
                step=1,
                wall_time=utils.get_timestamp_proto(),
            )
        ],
    )

    _TEST_TENSORBOARD_EXPERIMENT = tensorboard_experiment.TensorboardExperiment(
        name=_TEST_TENSORBOARD_EXPERIMENT_NAME,
        display_name=_TEST_DISPLAY_NAME,
    )


@dataclasses.dataclass(frozen=True)
class PipelineJobConstants:
    """Defines constants used by tests that create PipelineJob resources."""

    _TEST_PIPELINE_JOB_ID = "sample-test-pipeline-202111111"
    _TEST_PIPELINE_JOB_NAME = f"projects/{ProjectConstants._TEST_PROJECT}/locations/{ProjectConstants._TEST_LOCATION}/pipelineJobs/{_TEST_PIPELINE_JOB_ID}"
    _TEST_PIPELINE_CREATE_TIME = datetime.now()



# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1beta1.types import encryption_spec as gca_encryption_spec
from google.cloud.aiplatform_v1beta1.types import feature_monitoring_stats
from google.cloud.aiplatform_v1beta1.types import io
from google.cloud.aiplatform_v1beta1.types import job_state
from google.cloud.aiplatform_v1beta1.types import model_monitoring
from google.protobuf import duration_pb2  # type: ignore
from google.protobuf import struct_pb2  # type: ignore
from google.protobuf import timestamp_pb2  # type: ignore
from google.rpc import status_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1beta1",
    manifest={
        "ModelDeploymentMonitoringObjectiveType",
        "ModelDeploymentMonitoringJob",
        "ModelDeploymentMonitoringBigQueryTable",
        "ModelDeploymentMonitoringObjectiveConfig",
        "ModelDeploymentMonitoringScheduleConfig",
        "ModelMonitoringStatsAnomalies",
    },
)


class ModelDeploymentMonitoringObjectiveType(proto.Enum):
    r"""The Model Monitoring Objective types.

    Values:
        MODEL_DEPLOYMENT_MONITORING_OBJECTIVE_TYPE_UNSPECIFIED (0):
            Default value, should not be set.
        RAW_FEATURE_SKEW (1):
            Raw feature values' stats to detect skew
            between Training-Prediction datasets.
        RAW_FEATURE_DRIFT (2):
            Raw feature values' stats to detect drift
            between Serving-Prediction datasets.
        FEATURE_ATTRIBUTION_SKEW (3):
            Feature attribution scores to detect skew
            between Training-Prediction datasets.
        FEATURE_ATTRIBUTION_DRIFT (4):
            Feature attribution scores to detect skew
            between Prediction datasets collected within
            different time windows.
    """
    MODEL_DEPLOYMENT_MONITORING_OBJECTIVE_TYPE_UNSPECIFIED = 0
    RAW_FEATURE_SKEW = 1
    RAW_FEATURE_DRIFT = 2
    FEATURE_ATTRIBUTION_SKEW = 3
    FEATURE_ATTRIBUTION_DRIFT = 4


class ModelDeploymentMonitoringJob(proto.Message):
    r"""Represents a job that runs periodically to monitor the
    deployed models in an endpoint. It will analyze the logged
    training & prediction data to detect any abnormal behaviors.

    Attributes:
        name (str):
            Output only. Resource name of a
            ModelDeploymentMonitoringJob.
        display_name (str):
            Required. The user-defined name of the
            ModelDeploymentMonitoringJob. The name can be up
            to 128 characters long and can consist of any
            UTF-8 characters.
            Display name of a ModelDeploymentMonitoringJob.
        endpoint (str):
            Required. Endpoint resource name. Format:
            ``projects/{project}/locations/{location}/endpoints/{endpoint}``
        state (google.cloud.aiplatform_v1beta1.types.JobState):
            Output only. The detailed state of the
            monitoring job. When the job is still creating,
            the state will be 'PENDING'. Once the job is
            successfully created, the state will be
            'RUNNING'. Pause the job, the state will be
            'PAUSED'.
            Resume the job, the state will return to
            'RUNNING'.
        schedule_state (google.cloud.aiplatform_v1beta1.types.ModelDeploymentMonitoringJob.MonitoringScheduleState):
            Output only. Schedule state when the
            monitoring job is in Running state.
        latest_monitoring_pipeline_metadata (google.cloud.aiplatform_v1beta1.types.ModelDeploymentMonitoringJob.LatestMonitoringPipelineMetadata):
            Output only. Latest triggered monitoring
            pipeline metadata.
        model_deployment_monitoring_objective_configs (MutableSequence[google.cloud.aiplatform_v1beta1.types.ModelDeploymentMonitoringObjectiveConfig]):
            Required. The config for monitoring
            objectives. This is a per DeployedModel config.
            Each DeployedModel needs to be configured
            separately.
        model_deployment_monitoring_schedule_config (google.cloud.aiplatform_v1beta1.types.ModelDeploymentMonitoringScheduleConfig):
            Required. Schedule config for running the
            monitoring job.
        logging_sampling_strategy (google.cloud.aiplatform_v1beta1.types.SamplingStrategy):
            Required. Sample Strategy for logging.
        model_monitoring_alert_config (google.cloud.aiplatform_v1beta1.types.ModelMonitoringAlertConfig):
            Alert config for model monitoring.
        predict_instance_schema_uri (str):
            YAML schema file uri describing the format of
            a single instance, which are given to format
            this Endpoint's prediction (and explanation). If
            not set, we will generate predict schema from
            collected predict requests.
        sample_predict_instance (google.protobuf.struct_pb2.Value):
            Sample Predict instance, same format as
            [PredictRequest.instances][google.cloud.aiplatform.v1beta1.PredictRequest.instances],
            this can be set as a replacement of
            [ModelDeploymentMonitoringJob.predict_instance_schema_uri][google.cloud.aiplatform.v1beta1.ModelDeploymentMonitoringJob.predict_instance_schema_uri].
            If not set, we will generate predict schema from collected
            predict requests.
        analysis_instance_schema_uri (str):
            YAML schema file uri describing the format of a single
            instance that you want Tensorflow Data Validation (TFDV) to
            analyze.

            If this field is empty, all the feature data types are
            inferred from
            [predict_instance_schema_uri][google.cloud.aiplatform.v1beta1.ModelDeploymentMonitoringJob.predict_instance_schema_uri],
            meaning that TFDV will use the data in the exact format(data
            type) as prediction request/response. If there are any data
            type differences between predict instance and TFDV instance,
            this field can be used to override the schema. For models
            trained with Vertex AI, this field must be set as all the
            fields in predict instance formatted as string.
        bigquery_tables (MutableSequence[google.cloud.aiplatform_v1beta1.types.ModelDeploymentMonitoringBigQueryTable]):
            Output only. The created bigquery tables for
            the job under customer project. Customer could
            do their own query & analysis. There could be 4
            log tables in maximum:
            1. Training data logging predict
            request/response 2. Serving data logging predict
            request/response
        log_ttl (google.protobuf.duration_pb2.Duration):
            The TTL of BigQuery tables in user projects
            which stores logs. A day is the basic unit of
            the TTL and we take the ceil of TTL/86400(a
            day). e.g. { second: 3600} indicates ttl = 1
            day.
        labels (MutableMapping[str, str]):
            The labels with user-defined metadata to
            organize your ModelDeploymentMonitoringJob.

            Label keys and values can be no longer than 64
            characters (Unicode codepoints), can only
            contain lowercase letters, numeric characters,
            underscores and dashes. International characters
            are allowed.
            See https://goo.gl/xmQnxf for more information
            and examples of labels.
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Timestamp when this
            ModelDeploymentMonitoringJob was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Timestamp when this
            ModelDeploymentMonitoringJob was updated most
            recently.
        next_schedule_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Timestamp when this monitoring
            pipeline will be scheduled to run for the next
            round.
        stats_anomalies_base_directory (google.cloud.aiplatform_v1beta1.types.GcsDestination):
            Stats anomalies base folder path.
        encryption_spec (google.cloud.aiplatform_v1beta1.types.EncryptionSpec):
            Customer-managed encryption key spec for a
            ModelDeploymentMonitoringJob. If set, this
            ModelDeploymentMonitoringJob and all
            sub-resources of this
            ModelDeploymentMonitoringJob will be secured by
            this key.
        enable_monitoring_pipeline_logs (bool):
            If true, the scheduled monitoring pipeline logs are sent to
            Google Cloud Logging, including pipeline status and
            anomalies detected. Please note the logs incur cost, which
            are subject to `Cloud Logging
            pricing <https://cloud.google.com/logging#pricing>`__.
        error (google.rpc.status_pb2.Status):
            Output only. Only populated when the job's state is
            ``JOB_STATE_FAILED`` or ``JOB_STATE_CANCELLED``.
    """

    class MonitoringScheduleState(proto.Enum):
        r"""The state to Specify the monitoring pipeline.

        Values:
            MONITORING_SCHEDULE_STATE_UNSPECIFIED (0):
                Unspecified state.
            PENDING (1):
                The pipeline is picked up and wait to run.
            OFFLINE (2):
                The pipeline is offline and will be scheduled
                for next run.
            RUNNING (3):
                The pipeline is running.
        """
        MONITORING_SCHEDULE_STATE_UNSPECIFIED = 0
        PENDING = 1
        OFFLINE = 2
        RUNNING = 3

    class LatestMonitoringPipelineMetadata(proto.Message):
        r"""All metadata of most recent monitoring pipelines.

        Attributes:
            run_time (google.protobuf.timestamp_pb2.Timestamp):
                The time that most recent monitoring
                pipelines that is related to this run.
            status (google.rpc.status_pb2.Status):
                The status of the most recent monitoring
                pipeline.
        """

        run_time: timestamp_pb2.Timestamp = proto.Field(
            proto.MESSAGE,
            number=1,
            message=timestamp_pb2.Timestamp,
        )
        status: status_pb2.Status = proto.Field(
            proto.MESSAGE,
            number=2,
            message=status_pb2.Status,
        )

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    display_name: str = proto.Field(
        proto.STRING,
        number=2,
    )
    endpoint: str = proto.Field(
        proto.STRING,
        number=3,
    )
    state: job_state.JobState = proto.Field(
        proto.ENUM,
        number=4,
        enum=job_state.JobState,
    )
    schedule_state: MonitoringScheduleState = proto.Field(
        proto.ENUM,
        number=5,
        enum=MonitoringScheduleState,
    )
    latest_monitoring_pipeline_metadata: LatestMonitoringPipelineMetadata = proto.Field(
        proto.MESSAGE,
        number=25,
        message=LatestMonitoringPipelineMetadata,
    )
    model_deployment_monitoring_objective_configs: MutableSequence[
        "ModelDeploymentMonitoringObjectiveConfig"
    ] = proto.RepeatedField(
        proto.MESSAGE,
        number=6,
        message="ModelDeploymentMonitoringObjectiveConfig",
    )
    model_deployment_monitoring_schedule_config: "ModelDeploymentMonitoringScheduleConfig" = proto.Field(
        proto.MESSAGE,
        number=7,
        message="ModelDeploymentMonitoringScheduleConfig",
    )
    logging_sampling_strategy: model_monitoring.SamplingStrategy = proto.Field(
        proto.MESSAGE,
        number=8,
        message=model_monitoring.SamplingStrategy,
    )
    model_monitoring_alert_config: model_monitoring.ModelMonitoringAlertConfig = (
        proto.Field(
            proto.MESSAGE,
            number=15,
            message=model_monitoring.ModelMonitoringAlertConfig,
        )
    )
    predict_instance_schema_uri: str = proto.Field(
        proto.STRING,
        number=9,
    )
    sample_predict_instance: struct_pb2.Value = proto.Field(
        proto.MESSAGE,
        number=19,
        message=struct_pb2.Value,
    )
    analysis_instance_schema_uri: str = proto.Field(
        proto.STRING,
        number=16,
    )
    bigquery_tables: MutableSequence[
        "ModelDeploymentMonitoringBigQueryTable"
    ] = proto.RepeatedField(
        proto.MESSAGE,
        number=10,
        message="ModelDeploymentMonitoringBigQueryTable",
    )
    log_ttl: duration_pb2.Duration = proto.Field(
        proto.MESSAGE,
        number=17,
        message=duration_pb2.Duration,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=11,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=12,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=13,
        message=timestamp_pb2.Timestamp,
    )
    next_schedule_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=14,
        message=timestamp_pb2.Timestamp,
    )
    stats_anomalies_base_directory: io.GcsDestination = proto.Field(
        proto.MESSAGE,
        number=20,
        message=io.GcsDestination,
    )
    encryption_spec: gca_encryption_spec.EncryptionSpec = proto.Field(
        proto.MESSAGE,
        number=21,
        message=gca_encryption_spec.EncryptionSpec,
    )
    enable_monitoring_pipeline_logs: bool = proto.Field(
        proto.BOOL,
        number=22,
    )
    error: status_pb2.Status = proto.Field(
        proto.MESSAGE,
        number=23,
        message=status_pb2.Status,
    )


class ModelDeploymentMonitoringBigQueryTable(proto.Message):
    r"""ModelDeploymentMonitoringBigQueryTable specifies the BigQuery
    table name as well as some information of the logs stored in
    this table.

    Attributes:
        log_source (google.cloud.aiplatform_v1beta1.types.ModelDeploymentMonitoringBigQueryTable.LogSource):
            The source of log.
        log_type (google.cloud.aiplatform_v1beta1.types.ModelDeploymentMonitoringBigQueryTable.LogType):
            The type of log.
        bigquery_table_path (str):
            The created BigQuery table to store logs. Customer could do
            their own query & analysis. Format:
            ``bq://<project_id>.model_deployment_monitoring_<endpoint_id>.<tolower(log_source)>_<tolower(log_type)>``
    """

    class LogSource(proto.Enum):
        r"""Indicates where does the log come from.

        Values:
            LOG_SOURCE_UNSPECIFIED (0):
                Unspecified source.
            TRAINING (1):
                Logs coming from Training dataset.
            SERVING (2):
                Logs coming from Serving traffic.
        """
        LOG_SOURCE_UNSPECIFIED = 0
        TRAINING = 1
        SERVING = 2

    class LogType(proto.Enum):
        r"""Indicates what type of traffic does the log belong to.

        Values:
            LOG_TYPE_UNSPECIFIED (0):
                Unspecified type.
            PREDICT (1):
                Predict logs.
            EXPLAIN (2):
                Explain logs.
        """
        LOG_TYPE_UNSPECIFIED = 0
        PREDICT = 1
        EXPLAIN = 2

    log_source: LogSource = proto.Field(
        proto.ENUM,
        number=1,
        enum=LogSource,
    )
    log_type: LogType = proto.Field(
        proto.ENUM,
        number=2,
        enum=LogType,
    )
    bigquery_table_path: str = proto.Field(
        proto.STRING,
        number=3,
    )


class ModelDeploymentMonitoringObjectiveConfig(proto.Message):
    r"""ModelDeploymentMonitoringObjectiveConfig contains the pair of
    deployed_model_id to ModelMonitoringObjectiveConfig.

    Attributes:
        deployed_model_id (str):
            The DeployedModel ID of the objective config.
        objective_config (google.cloud.aiplatform_v1beta1.types.ModelMonitoringObjectiveConfig):
            The objective config of for the
            modelmonitoring job of this deployed model.
    """

    deployed_model_id: str = proto.Field(
        proto.STRING,
        number=1,
    )
    objective_config: model_monitoring.ModelMonitoringObjectiveConfig = proto.Field(
        proto.MESSAGE,
        number=2,
        message=model_monitoring.ModelMonitoringObjectiveConfig,
    )


class ModelDeploymentMonitoringScheduleConfig(proto.Message):
    r"""The config for scheduling monitoring job.

    Attributes:
        monitor_interval (google.protobuf.duration_pb2.Duration):
            Required. The model monitoring job scheduling
            interval. It will be rounded up to next full
            hour. This defines how often the monitoring jobs
            are triggered.
        monitor_window (google.protobuf.duration_pb2.Duration):
            The time window of the prediction data being included in
            each prediction dataset. This window specifies how long the
            data should be collected from historical model results for
            each run. If not set,
            [ModelDeploymentMonitoringScheduleConfig.monitor_interval][google.cloud.aiplatform.v1beta1.ModelDeploymentMonitoringScheduleConfig.monitor_interval]
            will be used. e.g. If currently the cutoff time is
            2022-01-08 14:30:00 and the monitor_window is set to be
            3600, then data from 2022-01-08 13:30:00 to 2022-01-08
            14:30:00 will be retrieved and aggregated to calculate the
            monitoring statistics.
    """

    monitor_interval: duration_pb2.Duration = proto.Field(
        proto.MESSAGE,
        number=1,
        message=duration_pb2.Duration,
    )
    monitor_window: duration_pb2.Duration = proto.Field(
        proto.MESSAGE,
        number=2,
        message=duration_pb2.Duration,
    )


class ModelMonitoringStatsAnomalies(proto.Message):
    r"""Statistics and anomalies generated by Model Monitoring.

    Attributes:
        objective (google.cloud.aiplatform_v1beta1.types.ModelDeploymentMonitoringObjectiveType):
            Model Monitoring Objective those stats and
            anomalies belonging to.
        deployed_model_id (str):
            Deployed Model ID.
        anomaly_count (int):
            Number of anomalies within all stats.
        feature_stats (MutableSequence[google.cloud.aiplatform_v1beta1.types.ModelMonitoringStatsAnomalies.FeatureHistoricStatsAnomalies]):
            A list of historical Stats and Anomalies
            generated for all Features.
    """

    class FeatureHistoricStatsAnomalies(proto.Message):
        r"""Historical Stats (and Anomalies) for a specific Feature.

        Attributes:
            feature_display_name (str):
                Display Name of the Feature.
            threshold (google.cloud.aiplatform_v1beta1.types.ThresholdConfig):
                Threshold for anomaly detection.
            training_stats (google.cloud.aiplatform_v1beta1.types.FeatureStatsAnomaly):
                Stats calculated for the Training Dataset.
            prediction_stats (MutableSequence[google.cloud.aiplatform_v1beta1.types.FeatureStatsAnomaly]):
                A list of historical stats generated by
                different time window's Prediction Dataset.
        """

        feature_display_name: str = proto.Field(
            proto.STRING,
            number=1,
        )
        threshold: model_monitoring.ThresholdConfig = proto.Field(
            proto.MESSAGE,
            number=3,
            message=model_monitoring.ThresholdConfig,
        )
        training_stats: feature_monitoring_stats.FeatureStatsAnomaly = proto.Field(
            proto.MESSAGE,
            number=4,
            message=feature_monitoring_stats.FeatureStatsAnomaly,
        )
        prediction_stats: MutableSequence[
            feature_monitoring_stats.FeatureStatsAnomaly
        ] = proto.RepeatedField(
            proto.MESSAGE,
            number=5,
            message=feature_monitoring_stats.FeatureStatsAnomaly,
        )

    objective: "ModelDeploymentMonitoringObjectiveType" = proto.Field(
        proto.ENUM,
        number=1,
        enum="ModelDeploymentMonitoringObjectiveType",
    )
    deployed_model_id: str = proto.Field(
        proto.STRING,
        number=2,
    )
    anomaly_count: int = proto.Field(
        proto.INT32,
        number=3,
    )
    feature_stats: MutableSequence[FeatureHistoricStatsAnomalies] = proto.RepeatedField(
        proto.MESSAGE,
        number=4,
        message=FeatureHistoricStatsAnomalies,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import MutableMapping, MutableSequence

import proto  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1beta1.schema.trainingjob.definition",
    manifest={
        "AutoMlImageObjectDetection",
        "AutoMlImageObjectDetectionInputs",
        "AutoMlImageObjectDetectionMetadata",
    },
)


class AutoMlImageObjectDetection(proto.Message):
    r"""A TrainingJob that trains and uploads an AutoML Image Object
    Detection Model.

    Attributes:
        inputs (google.cloud.aiplatform.v1beta1.schema.trainingjob.definition_v1beta1.types.AutoMlImageObjectDetectionInputs):
            The input parameters of this TrainingJob.
        metadata (google.cloud.aiplatform.v1beta1.schema.trainingjob.definition_v1beta1.types.AutoMlImageObjectDetectionMetadata):
            The metadata information
    """

    inputs: "AutoMlImageObjectDetectionInputs" = proto.Field(
        proto.MESSAGE,
        number=1,
        message="AutoMlImageObjectDetectionInputs",
    )
    metadata: "AutoMlImageObjectDetectionMetadata" = proto.Field(
        proto.MESSAGE,
        number=2,
        message="AutoMlImageObjectDetectionMetadata",
    )


class AutoMlImageObjectDetectionInputs(proto.Message):
    r"""

    Attributes:
        model_type (google.cloud.aiplatform.v1beta1.schema.trainingjob.definition_v1beta1.types.AutoMlImageObjectDetectionInputs.ModelType):

        budget_milli_node_hours (int):
            The training budget of creating this model, expressed in
            milli node hours i.e. 1,000 value in this field means 1 node
            hour. The actual metadata.costMilliNodeHours will be equal
            or less than this value. If further model training ceases to
            provide any improvements, it will stop without using the
            full budget and the metadata.successfulStopReason will be
            ``model-converged``. Note, node_hour = actual_hour \*
            number_of_nodes_involved. For modelType
            ``cloud``\ (default), the budget must be between 20,000 and
            900,000 milli node hours, inclusive. The default value is
            216,000 which represents one day in wall time, considering 9
            nodes are used. For model types ``mobile-tf-low-latency-1``,
            ``mobile-tf-versatile-1``, ``mobile-tf-high-accuracy-1`` the
            training budget must be between 1,000 and 100,000 milli node
            hours, inclusive. The default value is 24,000 which
            represents one day in wall time on a single node that is
            used.
        disable_early_stopping (bool):
            Use the entire training budget. This disables
            the early stopping feature. When false the early
            stopping feature is enabled, which means that
            AutoML Image Object Detection might stop
            training before the entire training budget has
            been used.
    """

    class ModelType(proto.Enum):
        r"""

        Values:
            MODEL_TYPE_UNSPECIFIED (0):
                Should not be set.
            CLOUD_HIGH_ACCURACY_1 (1):
                A model best tailored to be used within
                Google Cloud, and which cannot be exported.
                Expected to have a higher latency, but should
                also have a higher prediction quality than other
                cloud models.
            CLOUD_LOW_LATENCY_1 (2):
                A model best tailored to be used within
                Google Cloud, and which cannot be exported.
                Expected to have a low latency, but may have
                lower prediction quality than other cloud
                models.
            MOBILE_TF_LOW_LATENCY_1 (3):
                A model that, in addition to being available
                within Google Cloud can also be exported (see
                ModelService.ExportModel) and used on a mobile
                or edge device with TensorFlow afterwards.
                Expected to have low latency, but may have lower
                prediction quality than other mobile models.
            MOBILE_TF_VERSATILE_1 (4):
                A model that, in addition to being available
                within Google Cloud can also be exported (see
                ModelService.ExportModel) and used on a mobile
                or edge device with TensorFlow afterwards.
            MOBILE_TF_HIGH_ACCURACY_1 (5):
                A model that, in addition to being available
                within Google Cloud, can also be exported (see
                ModelService.ExportModel) and used on a mobile
                or edge device with TensorFlow afterwards.
                Expected to have a higher latency, but should
                also have a higher prediction quality than other
                mobile models.
        """
        MODEL_TYPE_UNSPECIFIED = 0
        CLOUD_HIGH_ACCURACY_1 = 1
        CLOUD_LOW_LATENCY_1 = 2
        MOBILE_TF_LOW_LATENCY_1 = 3
        MOBILE_TF_VERSATILE_1 = 4
        MOBILE_TF_HIGH_ACCURACY_1 = 5

    model_type: ModelType = proto.Field(
        proto.ENUM,
        number=1,
        enum=ModelType,
    )
    budget_milli_node_hours: int = proto.Field(
        proto.INT64,
        number=2,
    )
    disable_early_stopping: bool = proto.Field(
        proto.BOOL,
        number=3,
    )


class AutoMlImageObjectDetectionMetadata(proto.Message):
    r"""

    Attributes:
        cost_milli_node_hours (int):
            The actual training cost of creating this
            model, expressed in milli node hours, i.e. 1,000
            value in this field means 1 node hour.
            Guaranteed to not exceed
            inputs.budgetMilliNodeHours.
        successful_stop_reason (google.cloud.aiplatform.v1beta1.schema.trainingjob.definition_v1beta1.types.AutoMlImageObjectDetectionMetadata.SuccessfulStopReason):
            For successful job completions, this is the
            reason why the job has finished.
    """

    class SuccessfulStopReason(proto.Enum):
        r"""

        Values:
            SUCCESSFUL_STOP_REASON_UNSPECIFIED (0):
                Should not be set.
            BUDGET_REACHED (1):
                The inputs.budgetMilliNodeHours had been
                reached.
            MODEL_CONVERGED (2):
                Further training of the Model ceased to
                increase its quality, since it already has
                converged.
        """
        SUCCESSFUL_STOP_REASON_UNSPECIFIED = 0
        BUDGET_REACHED = 1
        MODEL_CONVERGED = 2

    cost_milli_node_hours: int = proto.Field(
        proto.INT64,
        number=1,
    )
    successful_stop_reason: SuccessfulStopReason = proto.Field(
        proto.ENUM,
        number=2,
        enum=SuccessfulStopReason,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-

# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from google.cloud.aiplatform.compat import services
from google.cloud.aiplatform.compat import types

V1BETA1 = "v1beta1"
V1 = "v1"

DEFAULT_VERSION = V1

if DEFAULT_VERSION == V1BETA1:

    services.dataset_service_client = services.dataset_service_client_v1beta1
    services.endpoint_service_client = services.endpoint_service_client_v1beta1
    services.featurestore_online_serving_service_client = (
        services.featurestore_online_serving_service_client_v1beta1
    )
    services.featurestore_service_client = services.featurestore_service_client_v1beta1
    services.job_service_client = services.job_service_client_v1beta1
    services.model_service_client = services.model_service_client_v1beta1
    services.pipeline_service_client = services.pipeline_service_client_v1beta1
    services.prediction_service_client = services.prediction_service_client_v1beta1
    services.specialist_pool_service_client = (
        services.specialist_pool_service_client_v1beta1
    )
    services.metadata_service_client = services.metadata_service_client_v1beta1
    services.tensorboard_service_client = services.tensorboard_service_client_v1beta1
    services.index_service_client = services.index_service_client_v1beta1
    services.index_endpoint_service_client = (
        services.index_endpoint_service_client_v1beta1
    )
    services.vizier_service_client = services.vizier_service_client_v1beta1

    types.accelerator_type = types.accelerator_type_v1beta1
    types.annotation = types.annotation_v1beta1
    types.annotation_spec = types.annotation_spec_v1beta1
    types.artifact = types.artifact_v1beta1
    types.batch_prediction_job = types.batch_prediction_job_v1beta1
    types.completion_stats = types.completion_stats_v1beta1
    types.context = types.context_v1beta1
    types.custom_job = types.custom_job_v1beta1
    types.data_item = types.data_item_v1beta1
    types.data_labeling_job = types.data_labeling_job_v1beta1
    types.dataset = types.dataset_v1beta1
    types.dataset_service = types.dataset_service_v1beta1
    types.deployed_model_ref = types.deployed_model_ref_v1beta1
    types.encryption_spec = types.encryption_spec_v1beta1
    types.endpoint = types.endpoint_v1beta1
    types.endpoint_service = types.endpoint_service_v1beta1
    types.entity_type = types.entity_type_v1beta1
    types.env_var = types.env_var_v1beta1
    types.event = types.event_v1beta1
    types.execution = types.execution_v1beta1
    types.explanation = types.explanation_v1beta1
    types.explanation_metadata = types.explanation_metadata_v1beta1
    types.feature = types.feature_v1beta1
    types.feature_monitoring_stats = types.feature_monitoring_stats_v1beta1
    types.feature_selector = types.feature_selector_v1beta1
    types.featurestore = types.featurestore_v1beta1
    types.featurestore_monitoring = types.featurestore_monitoring_v1beta1
    types.featurestore_online_service = types.featurestore_online_service_v1beta1
    types.featurestore_service = types.featurestore_service_v1beta1
    types.hyperparameter_tuning_job = types.hyperparameter_tuning_job_v1beta1
    types.index = types.index_v1beta1
    types.index_endpoint = types.index_endpoint_v1beta1
    types.io = types.io_v1beta1
    types.job_service = types.job_service_v1beta1
    types.job_state = types.job_state_v1beta1
    types.lineage_subgraph = types.lineage_subgraph_v1beta1
    types.machine_resources = types.machine_resources_v1beta1
    types.manual_batch_tuning_parameters = types.manual_batch_tuning_parameters_v1beta1
    types.matching_engine_deployed_index_ref = (
        types.matching_engine_deployed_index_ref_v1beta1
    )
    types.matching_engine_index = types.index_v1beta1
    types.matching_engine_index_endpoint = types.index_endpoint_v1beta1
    types.metadata_service = types.metadata_service_v1beta1
    types.metadata_schema = types.metadata_schema_v1beta1
    types.metadata_store = types.metadata_store_v1beta1
    types.model = types.model_v1beta1
    types.model_evaluation = types.model_evaluation_v1beta1
    types.model_evaluation_slice = types.model_evaluation_slice_v1beta1
    types.model_deployment_monitoring_job = (
        types.model_deployment_monitoring_job_v1beta1
    )
    types.model_monitoring = types.model_monitoring_v1beta1
    types.model_service = types.model_service_v1beta1
    types.operation = types.operation_v1beta1
    types.pipeline_failure_policy = types.pipeline_failure_policy_v1beta1
    types.pipeline_job = types.pipeline_job_v1beta1
    types.pipeline_service = types.pipeline_service_v1beta1
    types.pipeline_state = types.pipeline_state_v1beta1
    types.prediction_service = types.prediction_service_v1beta1
    types.specialist_pool = types.specialist_pool_v1beta1
    types.specialist_pool_service = types.specialist_pool_service_v1beta1
    types.study = types.study_v1beta1
    types.tensorboard = types.tensorboard_v1beta1
    types.tensorboard_service = types.tensorboard_service_v1beta1
    types.tensorboard_data = types.tensorboard_data_v1beta1
    types.tensorboard_experiment = types.tensorboard_experiment_v1beta1
    types.tensorboard_run = types.tensorboard_run_v1beta1
    types.tensorboard_service = types.tensorboard_service_v1beta1
    types.tensorboard_time_series = types.tensorboard_time_series_v1beta1
    types.training_pipeline = types.training_pipeline_v1beta1
    types.types = types.types_v1beta1
    types.vizier_service = types.vizier_service_v1beta1

if DEFAULT_VERSION == V1:

    services.dataset_service_client = services.dataset_service_client_v1
    services.endpoint_service_client = services.endpoint_service_client_v1
    services.featurestore_online_serving_service_client = (
        services.featurestore_online_serving_service_client_v1
    )
    services.featurestore_service_client = services.featurestore_service_client_v1
    services.job_service_client = services.job_service_client_v1
    services.model_service_client = services.model_service_client_v1
    services.pipeline_service_client = services.pipeline_service_client_v1
    services.prediction_service_client = services.prediction_service_client_v1
    services.specialist_pool_service_client = services.specialist_pool_service_client_v1
    services.tensorboard_service_client = services.tensorboard_service_client_v1
    services.index_service_client = services.index_service_client_v1
    services.index_endpoint_service_client = services.index_endpoint_service_client_v1
    services.vizier_service_client = services.vizier_service_client_v1

    types.accelerator_type = types.accelerator_type_v1
    types.annotation = types.annotation_v1
    types.annotation_spec = types.annotation_spec_v1
    types.artifact = types.artifact_v1
    types.batch_prediction_job = types.batch_prediction_job_v1
    types.completion_stats = types.completion_stats_v1
    types.context = types.context_v1
    types.custom_job = types.custom_job_v1
    types.data_item = types.data_item_v1
    types.data_labeling_job = types.data_labeling_job_v1
    types.dataset = types.dataset_v1
    types.dataset_service = types.dataset_service_v1
    types.deployed_model_ref = types.deployed_model_ref_v1
    types.encryption_spec = types.encryption_spec_v1
    types.endpoint = types.endpoint_v1
    types.endpoint_service = types.endpoint_service_v1
    types.entity_type = types.entity_type_v1
    types.env_var = types.env_var_v1
    types.event = types.event_v1
    types.execution = types.execution_v1
    types.explanation = types.explanation_v1
    types.explanation_metadata = types.explanation_metadata_v1
    types.feature = types.feature_v1
    types.feature_monitoring_stats = types.feature_monitoring_stats_v1
    types.feature_selector = types.feature_selector_v1
    types.featurestore = types.featurestore_v1
    types.featurestore_online_service = types.featurestore_online_service_v1
    types.featurestore_service = types.featurestore_service_v1
    types.hyperparameter_tuning_job = types.hyperparameter_tuning_job_v1
    types.index = types.index_v1
    types.index_endpoint = types.index_endpoint_v1
    types.io = types.io_v1
    types.job_service = types.job_service_v1
    types.job_state = types.job_state_v1
    types.lineage_subgraph = types.lineage_subgraph_v1
    types.machine_resources = types.machine_resources_v1
    types.manual_batch_tuning_parameters = types.manual_batch_tuning_parameters_v1
    types.matching_engine_deployed_index_ref = (
        types.matching_engine_deployed_index_ref_v1
    )
    types.matching_engine_index = types.index_v1
    types.matching_engine_index_endpoint = types.index_endpoint_v1
    types.metadata_service = types.metadata_service_v1
    types.metadata_schema = types.metadata_schema_v1
    types.metadata_store = types.metadata_store_v1
    types.model = types.model_v1
    types.model_evaluation = types.model_evaluation_v1
    types.model_evaluation_slice = types.model_evaluation_slice_v1
    types.model_deployment_monitoring_job = types.model_deployment_monitoring_job_v1
    types.model_monitoring = types.model_monitoring_v1
    types.model_service = types.model_service_v1
    types.operation = types.operation_v1
    types.pipeline_failure_policy = types.pipeline_failure_policy_v1
    types.pipeline_job = types.pipeline_job_v1
    types.pipeline_service = types.pipeline_service_v1
    types.pipeline_state = types.pipeline_state_v1
    types.prediction_service = types.prediction_service_v1
    types.specialist_pool = types.specialist_pool_v1
    types.specialist_pool_service = types.specialist_pool_service_v1
    types.study = types.study_v1
    types.tensorboard = types.tensorboard_v1
    types.tensorboard_service = types.tensorboard_service_v1
    types.tensorboard_data = types.tensorboard_data_v1
    types.tensorboard_experiment = types.tensorboard_experiment_v1
    types.tensorboard_run = types.tensorboard_run_v1
    types.tensorboard_service = types.tensorboard_service_v1
    types.tensorboard_time_series = types.tensorboard_time_series_v1
    types.training_pipeline = types.training_pipeline_v1
    types.types = types.types_v1
    types.vizier_service = types.vizier_service_v1

__all__ = (
    DEFAULT_VERSION,
    V1BETA1,
    V1,
    services,
    types,
)
